"""Review stage."""

from dataclasses import dataclass, field
import json
from typing import Any, Dict, Optional

from redbrick.common.stage import Stage


@dataclass
class ReviewStage(Stage):
    """Review Stage.

    Parameters
    --------------
    stage_name: str
        Stage name.

    on_accept: Optional[str] = None
        The next stage for the task when accepted in current stage.
        If None, will go to Output stage.

    on_reject: Optional[str] = None
        The next stage for the task when rejected in current stage.
        If None, will go to Output Stage.

    config: Config
        Stage config.
    """

    @dataclass
    class Config(Stage.Config):
        """Review Stage Config.

        Parameters
        --------------
        review_percentage: Optional[float]
            Percentage of tasks in [0, 1] that will be sampled for review. (Default: 1)

        auto_assignment: Optional[bool]
            Enable task auto assignment. (Default: True)

        auto_assignment_queue_size: Optional[int]
            Task auto-assignment queue size. (Default: 5)
        """

        review_percentage: Optional[float] = None
        auto_assignment: Optional[bool] = None
        auto_assignment_queue_size: Optional[int] = None

        @classmethod
        def from_entity(cls, entity: Optional[Dict] = None) -> "ReviewStage.Config":
            """Get object from entity."""
            if not entity:
                return cls()
            return cls(
                review_percentage=entity.get("reviewPercent"),
                auto_assignment=entity.get("autoAssign"),
                auto_assignment_queue_size=entity.get("queueSize"),
            )

        def to_entity(self) -> Dict:
            """Get entity from object."""
            entity: Dict[str, Any] = {}
            if self.review_percentage is not None:
                entity["reviewPercent"] = self.review_percentage
            if self.auto_assignment is not None:
                entity["autoAssign"] = self.auto_assignment
            if self.auto_assignment_queue_size is not None:
                entity["queueSize"] = self.auto_assignment_queue_size
            return entity

    stage_name: str
    on_accept: Optional[str] = None
    on_reject: Optional[str] = None
    config: Config = field(default_factory=Config.from_entity)

    @classmethod
    def from_entity(cls, entity: Dict) -> "ReviewStage":
        """Get object from entity"""
        config = entity.get("stageConfig")
        if config and isinstance(config, str):
            config = json.loads(config)
        return cls(
            stage_name=entity["stageName"],
            on_accept=entity["routing"]["passed"],
            on_reject=entity["routing"]["failed"],
            config=cls.Config.from_entity(config or {}),
        )

    def to_entity(self) -> Dict:
        """Get entity from object."""
        return {
            "brickName": "expert-review",
            "stageName": self.stage_name,
            "routing": {
                "passed": self.on_accept or "Output",
                "failed": self.on_reject or "Output",
            },
            "stageConfig": self.config.to_entity(),
        }



"""Handlers to access APIs for project settings."""

from typing import Dict

from redbrick.common.client import RBClient
from redbrick.common.settings import (
    SettingsControllerInterface,
    LabelValidation,
    HangingProtocol,
)


class SettingsRepo(SettingsControllerInterface):
    """Class to manage interaction with project settings APIs."""

    def __init__(self, client: RBClient) -> None:
        """Construct SettingsRepo."""
        self.client = client

    def get_label_validation(self, org_id: str, project_id: str) -> LabelValidation:
        """Get project label validation setting."""
        query = """
            query getLabelValidationSDK($orgId: UUID!, $projectId: UUID!) {
                project(orgId: $orgId, projectId: $projectId) {
                    labelValidationSettings {
                        enabled
                        enforce
                        script
                    }
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        if response.get("project"):
            return response["project"]["labelValidationSettings"]

        raise Exception("Project does not exist")

    def set_label_validation(
        self, org_id: str, project_id: str, label_validation: LabelValidation
    ) -> None:
        """Set project label validation setting."""
        query = """
            mutation setLabelValidationSDK(
                $orgId: UUID!
                $projectId: UUID!
                $enabled: Boolean!
                $enforce: Boolean!
                $script: String
            ) {
                updateLabelValidationSettings(
                    orgId: $orgId
                    projectId: $projectId
                    enabled: $enabled
                    enforce: $enforce
                    script: $script
                ) {
                    ok
                }
            }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "enabled": label_validation.get("enabled", False) or False,
            "enforce": label_validation.get("enforce", False) or False,
            "script": label_validation.get("script"),
        }
        self.client.execute_query(query, variables)

    def get_hanging_protocol(self, org_id: str, project_id: str) -> HangingProtocol:
        """Get project hanging protocol setting."""
        query = """
            query getHangingProtocolSDK($orgId: UUID!, $projectId: UUID!) {
                project(orgId: $orgId, projectId: $projectId) {
                    hangingProtocol {
                        enabled
                        script
                    }
                }
            }
        """
        variables = {"orgId": org_id, "projectId": project_id}
        response: Dict[str, Dict] = self.client.execute_query(query, variables)
        if response.get("project"):
            return response["project"]["hangingProtocol"]

        raise Exception("Project does not exist")

    def set_hanging_protocol(
        self, org_id: str, project_id: str, hanging_protocol: HangingProtocol
    ) -> None:
        """Set project hanging protocol setting."""
        query = """
            mutation setHangingProtocolSDK(
                $orgId: UUID!
                $projectId: UUID!
                $enabled: Boolean!
                $script: String
            ) {
                updateHangingProtocolScript(
                    orgId: $orgId
                    projectId: $projectId
                    enabled: $enabled
                    script: $script
                ) {
                    ok
                }
            }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "enabled": hanging_protocol.get("enabled", False) or False,
            "script": hanging_protocol.get("script"),
        }
        self.client.execute_query(query, variables)

    def toggle_reference_standard_task(
        self, org_id: str, project_id: str, task_id: str, enable: bool
    ) -> None:
        """Toggle reference standard task."""
        query = """
            mutation toggleReferenceStandardTaskSDK(
                $orgId: UUID!
                $projectId: UUID!
                $taskId: UUID!
                $enable: Boolean!
            ) {
                toggleReferenceStandardTask(
                    orgId: $orgId
                    projectId: $projectId
                    taskId: $taskId
                    enable: $enable
                ) {
                    ok
                }
            }
        """
        variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskId": task_id,
            "enable": enable,
        }
        self.client.execute_query(query, variables)



"""A utility iterator to handle default RedBrick pagination behavior."""

from typing import Any, Dict, List, Optional, Callable, Tuple


class PaginationIterator:
    """Construct Labelset Iterator."""

    def __init__(
        self,
        func: Callable[[int, Optional[str]], Tuple[List[Dict], Optional[str]]],
        concurrency: int = 10,
        limit: Optional[int] = None,
    ) -> None:
        """Construct LabelsetIterator."""
        self.cursor: Optional[str] = None
        self.datapoints_batch: Optional[List[Dict]] = None
        self.datapoints_batch_index: Optional[int] = None

        self.func = func
        self.concurrency = concurrency
        self.limit = limit

        self.total = 0

    def __iter__(self) -> Any:
        """Get iterator."""
        return self

    def __len__(self) -> int:
        """Get length of iteration."""
        return self.total

    def __next__(self) -> Dict:
        """Get next batch of labels / datapoint."""
        # If cursor is None and current datapoints_batch has been processed
        if (
            self.datapoints_batch_index is not None
            and self.cursor is None
            and self.datapoints_batch
            and len(self.datapoints_batch) == self.datapoints_batch_index
        ):
            raise StopIteration

        # If current datapoints_batch is None or we have finished
        # processing current datapoints_batch
        if (
            self.datapoints_batch is None
            or len(self.datapoints_batch) == self.datapoints_batch_index
        ):
            # When no data is returned in the current iteration,
            # but there is still more data, go for the next iteration
            while True:
                self.datapoints_batch, self.cursor, *_ = self.func(
                    (
                        max(0, min(self.concurrency, self.limit - self.total))
                        if self.limit is not None
                        else self.concurrency
                    ),
                    self.cursor,
                )
                if (
                    self.limit is not None
                    and self.total + len(self.datapoints_batch) >= self.limit
                ):
                    self.datapoints_batch = self.datapoints_batch[
                        : max(0, self.limit - self.total)
                    ]
                    self.cursor = None
                if not self.datapoints_batch and self.cursor:
                    continue
                self.datapoints_batch_index = 0
                self.total += len(self.datapoints_batch)
                break

        # Current entry to return
        if self.datapoints_batch and self.datapoints_batch_index is not None:
            entry = self.datapoints_batch[self.datapoints_batch_index]
            self.datapoints_batch_index += 1

            return entry

        raise StopIteration



"""Partial queries to prevent duplication."""

PROJECT_SHARD = """
orgId
projectId
name
desc
status
tdType
taxonomy {
    name
}
projectUrl
createdAt
consensusSettings {
    enabled
}
workspace {
    workspaceId
}
"""

STAGE_SHARD = """
stageName
brickName
stageConfig
routing {
    ...on NoRouting {
        placeholder
    }
    ...on NextRouting {
        nextStageName
    }
    ...on MultiRouting {
        stageNames
    }
    ...on BooleanRouting {
        passed
        failed
    }
    ...on FeedbackRouting {
        feedbackStageName
    }
}
"""

TAXONOMY_SHARD = """
orgId
name
version
createdAt
categories {
    name
    children {
        name
        classId
        disabled
        children {
            name
            classId
            disabled
            children {
                name
                classId
                disabled
                children {
                    name
                    classId
                    disabled
                    children {
                        name
                        classId
                        disabled
                        children {
                            name
                            classId
                            disabled
                        }
                    }
                }
            }
        }
    }
}
attributes {
    name
    attrType
    whitelist
    disabled
}
taskCategories {
    name
    children {
        name
        classId
        disabled
        children {
            name
            classId
            disabled
            children {
                name
                classId
                disabled
                children {
                    name
                    classId
                    disabled
                    children {
                        name
                        classId
                        disabled
                        children {
                            name
                            classId
                            disabled
                        }
                    }
                }
            }
        }
    }
}
taskAttributes {
    name
    attrType
    whitelist
    disabled
}
colorMap {
    name
    color
    classid
    trail
    taskcategory
}
archived
isNew
taxId
studyClassify {
    name
    attrType
    attrId
    options {
        name
        optionId
        color
        archived
    }
    archived
    parents
    hint
    defaultValue
}
seriesClassify {
    name
    attrType
    attrId
    options {
        name
        optionId
        color
        archived
    }
    archived
    parents
    hint
    defaultValue
}
instanceClassify {
    name
    attrType
    attrId
    options {
        name
        optionId
        color
        archived
    }
    archived
    parents
    hint
    defaultValue
}
objectTypes {
    category
    classId
    labelType
    attributes {
        name
        attrType
        attrId
        options {
            name
            optionId
            color
            archived
        }
        archived
        parents
        hint
        defaultValue
    }
    color
    archived
    parents
    hint
}
"""

TASK_DATA_SHARD = """
    createdAt
    createdByEmail
    labelsData(interpolate: true)
    labelsStorage {
        storageId
    }
    labelsMap(presigned: false) {
        imageIndex
        labelName
    }
"""

TASK_SHARD = f"""
    ... on LabelingTask {{
        state
        assignedTo {{
            userId
            email
        }}
        taskData {{
            {TASK_DATA_SHARD}
        }}
        subTasks {{
            state
            assignedTo {{
                userId
                email
            }}
            taskData {{
                {TASK_DATA_SHARD}
            }}
        }}
        overallConsensusScore
        consensusInfo {{
            user {{
                userId
                email
            }}
            taskData {{
                {TASK_DATA_SHARD}
            }}
            scores {{
                user {{
                    userId
                    email
                }}
                score
            }}
        }}
    }}
"""


def datapoint_shard(raw_items: bool, presigned_items: bool) -> str:
    """Return the datapoint shard."""
    return f"""
        name
        {"items(presigned: false)" if raw_items else ""}
        {"itemsPresigned:items(presigned: true)" if presigned_items else ""}
        createdAt
        createdByEntity {{
            userId
            email
        }}
        metaData
        seriesInfo {{
            name
            itemsIndices
            dataType
            metaData
        }}
        storageMethod {{
            storageId
        }}
    """


def task_shard(presigned_items: bool, with_consensus: bool) -> str:
    """Return the task shard for the router query."""
    return f"""
        taskId
        dpId
        currentStageName
        priority
        currentStageSubTask{"(consensus: true)" if with_consensus else ""} {{
            {TASK_SHARD}
        }}
        datapoint {{
            {datapoint_shard(True, presigned_items)}
        }}
        latestTaskData {{
            {TASK_DATA_SHARD}
        }}
    """


def router_task_shard(with_labels: bool) -> str:
    """Return router task shard for events query."""
    return f"""
        taskId
        currentStageName
        datapoint {{
            {datapoint_shard(True, False)}
        }}
        priority
        genericEvents {{
            __typename
            ... on TaskEvent {{
                eventId
                createdAt
                createEvent {{
                    currentStageName
                    isGroundTruth
                    priority
                }}
                inputEvent {{
                    currentStageName
                    overallConsensusScore
                    priority
                }}
                outputEvent {{
                    currentStageName
                    outputBool
                    timeSpentMs
                }}
                taskData {{
                    stageName
                    createdBy
                    {TASK_DATA_SHARD if with_labels else ''}
                }}
            }}
            ... on Comment {{
                commentId
                createdBy {{
                    userId
                }}
                textVal
                createdAt
                stageName
                issueComment
                issueResolved
                replies {{
                    commentId
                    createdBy {{
                        userId
                    }}
                    textVal
                    createdAt
                    stageName
                    issueComment
                    issueResolved
                }}
            }}
            ... on TaskStateChanges {{
                stageNameAfter: stageName
                assignedAtAfter
                createdAt
                statusBefore
                statusAfter
                assignedToBefore
                assignedToAfter
                consensusAssigneesBefore
                consensusAssigneesAfter
                consensusStatusesBefore
                consensusStatusesAfter
                reviewResultBefore
                reviewResultAfter
            }}
        }}
    """



"""Utilities for working with label objects."""

import os
from typing import Any, Dict, List, Optional, Union
import json
from copy import deepcopy

from redbrick.stage import ReviewStage
from redbrick.types import task as TaskType


def clean_rb_label(label: Dict) -> Dict:
    """Clean any None fields."""
    for key, val in label.copy().items():
        if val is None:
            del label[key]
    return label


def user_format(user: Optional[str], users: Dict[str, str]) -> Optional[str]:
    """User format."""
    if not user:
        return user
    if user.startswith("RB:"):
        return "System"
    if user.startswith("API:"):
        return "API Key"
    return users.get(user, user)


def assignee_format(
    task: Optional[Dict[str, Any]], users: Dict[str, str]
) -> Optional[Dict]:
    """Assignee format."""
    if not task:
        return None

    user_val = user_format((task.get("assignedTo", {}) or {}).get("userId"), users)
    if not user_val:
        return None

    assignee = {"user": user_val, "status": task["state"]}
    if task.get("assignedAt"):
        assignee["assignedAt"] = task["assignedAt"]
    if task.get("progressSavedAt"):
        assignee["lastSavedAt"] = task["progressSavedAt"]
    if task.get("completedAt"):
        assignee["completedAt"] = task["completedAt"]
    if task.get("completionTimeMs") is not None:
        assignee["timeSpentMs"] = task["completionTimeMs"]

    return assignee


def from_rb_task_data(task_data: Dict) -> Dict:
    """Get object from task data."""
    return {
        "updatedAt": task_data.get("createdAt"),
        "labels": [
            clean_rb_label(label)
            for label in json.loads(task_data.get("labelsData", "[]") or "[]")
        ],
        "labelsMap": task_data.get("labelsMap", []) or [],
        "labelStorageId": (task_data.get("labelsStorage", {}) or {}).get("storageId"),
    }


def from_rb_sub_task(task: Dict) -> Dict:
    """Get task object from sub task."""
    assignee = task.get("assignedTo", {}) or {}
    return {
        "status": task.get("state"),
        "assignee": user_format(
            assignee.get("userId"),
            {
                assignee.get("userId", ""): assignee.get(
                    "email", assignee.get("userId", "")
                )
            },
        ),
        **from_rb_task_data(task.get("taskData", {}) or {}),
    }


def from_rb_consensus_info(task: Dict) -> Dict:
    """Get consensus info from task."""
    scores = []
    for task_score in task.get("scores", []) or []:
        assignee = task_score["user"]
        score: Dict[str, Any] = {
            "assignee": user_format(
                assignee.get("userId"),
                {
                    assignee.get("userId", ""): assignee.get(
                        "email", assignee.get("userId", "")
                    )
                },
            )
        }
        score["score"] = (task_score.get("score", 0) or 0) * 100
        scores.append(score)

    assignee = task.get("user", {}) or {}
    return {
        "assignee": user_format(
            assignee.get("userId"),
            {
                assignee.get("userId", ""): assignee.get(
                    "email", assignee.get("userId", "")
                )
            },
        ),
        **from_rb_task_data(task.get("taskData", {}) or {}),
        "scores": scores,
    }


def flat_rb_format(
    labels: List[Dict],
    items: List[str],
    items_presigned: List[str],
    name: str,
    created_by: Optional[str],
    created_at: str,
    updated_by: Optional[str],
    updated_at: Optional[str],
    task_id: str,
    current_stage_name: str,
    priority: Optional[float],
    labels_map: List[Optional[Dict]],
    series_info: Optional[List[Dict]],
    meta_data: Optional[Dict],
    storage_id: str,
    label_storage_id: Optional[str],
    current_stage_sub_task: Optional[Dict],
) -> Dict:
    """Get standard rb flat format, same as import format."""
    # pylint: disable=too-many-locals
    task: Dict[str, Any] = {
        "taskId": task_id,
        "name": name,
        "items": items,
        "itemsPresigned": items_presigned,
        "currentStageName": current_stage_name,
        "createdBy": created_by,
        "createdAt": created_at,
        "updatedBy": updated_by,
        "updatedAt": updated_at,
        "labels": labels,
        "labelsMap": labels_map,
        "seriesInfo": series_info,
        "metaData": meta_data,
        "storageId": storage_id,
        "labelStorageId": label_storage_id,
        "priority": priority,
    }

    if current_stage_sub_task:
        if current_stage_sub_task.get("subTasks"):
            task["consensusTasks"] = [from_rb_sub_task(current_stage_sub_task)]
            for sub_task in current_stage_sub_task["subTasks"]:
                task["consensusTasks"].append(from_rb_sub_task(sub_task))

        elif len(current_stage_sub_task.get("consensusInfo", []) or []) > 1:
            task["consensusScore"] = (
                current_stage_sub_task.get("overallConsensusScore", 0) or 0
            ) * 100

            task["consensusTasks"] = []
            for consensus_task in current_stage_sub_task["consensusInfo"]:
                task["consensusTasks"].append(from_rb_consensus_info(consensus_task))

    return task


def parse_entry_latest(item: Dict) -> Dict:
    """Parse entry latest."""
    try:
        task_id = item["taskId"]
        task_data = item["latestTaskData"] or {}
        datapoint = item["datapoint"]
        items = datapoint["items"]
        items_presigned = datapoint.get("itemsPresigned", []) or []
        name = datapoint["name"]
        created_by = (datapoint.get("createdByEntity", {}) or {}).get("email")
        created_at = datapoint["createdAt"]
        updated_by = task_data.get("createdByEmail")
        updated_at = task_data.get("createdAt")
        labels = [
            clean_rb_label(label)
            for label in json.loads(task_data.get("labelsData") or "[]")
        ]
        storage_id = datapoint["storageMethod"]["storageId"]
        label_storage_id = (task_data.get("labelsStorage") or {}).get("storageId")

        return flat_rb_format(
            labels,
            items,
            items_presigned,
            name,
            created_by,
            created_at,
            updated_by,
            updated_at,
            task_id,
            item["currentStageName"],
            item["priority"],
            task_data.get("labelsMap", []) or [],
            datapoint.get("seriesInfo"),
            json.loads(datapoint["metaData"]) if datapoint.get("metaData") else None,
            storage_id,
            label_storage_id,
            item.get("currentStageSubTask"),
        )
    except (AttributeError, KeyError, TypeError, json.decoder.JSONDecodeError):
        return {}


def dicom_rb_series(
    item_index_map: Dict[int, int],
    input_task: Dict,
    output_task: TaskType.OutputTask,
    taxonomy: Dict,
) -> None:
    """Get standard rb flat format, same as import format."""
    # pylint: disable=too-many-branches, too-many-statements, too-many-locals
    labels = input_task.get("labels", []) or []
    labels_map = input_task.get("labelsMap", []) or []
    series = output_task["series"]

    segmentation_mapping: Dict[int, Dict[Optional[str], Union[str, List[str]]]] = {}
    for idx, label_map in enumerate(labels_map):
        volume_index = (
            item_index_map[label_map["imageIndex"]]
            if "imageIndex" in label_map and label_map["imageIndex"] in item_index_map
            else idx
        )
        segmentation_mapping[volume_index] = {}

        if volume_index >= len(series):
            series.extend([{} for _ in range(volume_index - len(series) + 1)])
        if label_map and label_map.get("labelName"):
            series[volume_index]["segmentations"] = label_map["labelName"]
            series[volume_index]["segmentMap"] = {}
            series[volume_index]["binaryMask"] = (
                label_map.get("binaryMask", False) or False
            )
            series[volume_index]["semanticMask"] = (
                label_map.get("semanticMask", False) or False
            )
            series[volume_index]["pngMask"] = label_map.get("pngMask", False) or False

            segmentations = series[volume_index].get("segmentations")
            if isinstance(segmentations, str):
                segmentation_mapping[volume_index][None] = segmentations
            elif series[volume_index].get("binaryMask") and isinstance(
                segmentations, list
            ):
                for segmentation in segmentations:
                    segmentation_mapping[volume_index][
                        os.path.basename(segmentation)
                    ] = segmentation

    for label in labels:
        volume_index = label.get("volumeindex", 0)
        if volume_index >= len(series):
            series.extend([{} for _ in range(volume_index - len(series) + 1)])

    for label in labels:
        volume: TaskType.Series = series[label.get("volumeindex", 0)]
        label_obj: TaskType.CommonLabelProps = {}
        if "category" in label and label["category"] is not None:
            label_obj["category"] = (
                label["category"]
                if not isinstance(label["category"], list)
                else (
                    label["category"][0][1]
                    if isinstance(label["category"][0], list)
                    and len(label["category"][0]) == 2
                    else (
                        label["category"][0][1:]
                        if isinstance(label["category"][0], list)
                        and len(label["category"][0]) > 2
                        else label["category"]
                    )
                )
            )

        attributes = {}
        for attribute in label.get("attributes", []) or []:
            attributes[attribute["name"]] = (
                attribute["value"]
                if not isinstance(attribute["value"], str)
                else (
                    True
                    if attribute["value"].lower() == "true"
                    else (
                        False
                        if attribute["value"].lower() == "false"
                        else attribute["value"]
                    )
                )
            )
        if attributes:
            label_obj["attributes"] = attributes

        video_metadata: Dict[str, TaskType.VideoMetaData] = {}
        if isinstance(label.get("frameindex"), int) and label["frameindex"] >= 0:
            video_metadata = {
                "video": {
                    "frameIndex": label["frameindex"],
                    "trackId": label.get("trackid", ""),
                    "keyFrame": label.get("keyframe", True),
                    "endTrack": label.get("end", True),
                }
            }

        items: List[str] = volume.get("items", []) or []  # type: ignore

        measurement_stats = {}
        if isinstance(label.get("stats"), dict):
            measurement_stats = {
                "stats": {
                    prop: val for prop, val in label["stats"].items() if val is not None
                }
            }

        if label.get("tasklevelclassify") or label.get("studyclassify"):
            output_task["classification"] = {**label_obj}  # type: ignore
        elif label.get("multiclassify") or label.get("seriesclassify"):
            volume["classifications"] = volume.get("classifications", []) or []
            volume["classifications"].append(
                {
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                }
            )
        elif label.get("instanceclassify"):
            volume["instanceClassifications"] = (
                volume.get("instanceClassifications", []) or []
            )
            volume["instanceClassifications"].append(
                {
                    "fileIndex": label["frameindex"],
                    "fileName": (
                        items[label["frameindex"]]
                        if label["frameindex"] < len(items)
                        else ""
                    ),
                    "values": label_obj.get("attributes", {}),
                }
            )
        elif label.get("dicom", {}).get("instanceid"):
            volume_indices = (
                [label["volumeindex"]]
                if isinstance(label.get("volumeindex"), int)
                and 0 <= label["volumeindex"] < len(series)
                else list(range(len(series)))
            )
            for volume_index in volume_indices:
                series[volume_index]["segmentMap"] = series[volume_index].get(
                    "segmentMap", {}
                )
                series[volume_index]["binaryMask"] = (
                    series[volume_index].get("binaryMask", False) or False
                )
                series[volume_index]["semanticMask"] = (
                    series[volume_index].get("semanticMask", False) or False
                )
                series[volume_index]["pngMask"] = (
                    series[volume_index].get("pngMask", False) or False
                )
                instance: Optional[str] = None
                if bool(taxonomy.get("isNew")) and series[volume_index].get(
                    "semanticMask"
                ):
                    instance = (
                        str(label["classid"] + 1)
                        if label["classid"] + 1
                        not in series[volume_index].get("segmentMap", {})
                        else None
                    )
                else:
                    instance = str(label["dicom"]["instanceid"])
                if instance is not None:
                    mask = (segmentation_mapping.get(volume_index, {}) or {}).get(
                        f"mask-{instance}.png"
                        if series[volume_index].get("pngMask")
                        else (
                            f"category-{instance}.nii.gz"
                            if series[volume_index].get("semanticMask")
                            else (
                                f"instance-{instance}.nii.gz"
                                if series[volume_index].get("binaryMask")
                                else None
                            )
                        )
                    )
                    segment_map_instance: TaskType.CommonLabelProps = {**label_obj}
                    if mask:
                        segment_map_instance["mask"] = mask

                    cur_series = series[volume_index]
                    if "segmentMap" not in cur_series:
                        cur_series["segmentMap"] = {}
                    cur_series["segmentMap"][instance] = segment_map_instance

        elif label.get("length3d"):
            volume["measurements"] = volume.get("measurements", [])
            volume["measurements"].append(
                {
                    "type": "length",
                    "point1": dict(zip("ijk", label["length3d"]["point1"])),
                    "point2": dict(zip("ijk", label["length3d"]["point2"])),
                    "absolutePoint1": dict(
                        zip("xyz", label["length3d"]["computedpoint1world"])
                    ),
                    "absolutePoint2": dict(
                        zip("xyz", label["length3d"]["computedpoint2world"])
                    ),
                    "normal": label["length3d"]["normal"],
                    "length": label["length3d"]["computedlength"],
                    **label_obj,  # type: ignore
                }
            )
        elif label.get("angle3d"):
            volume["measurements"] = volume.get("measurements", [])
            volume["measurements"].append(
                {
                    "type": "angle",
                    "point1": dict(zip("ijk", label["angle3d"]["point1"])),
                    "vertex": dict(zip("ijk", label["angle3d"]["point2"])),
                    "point2": dict(zip("ijk", label["angle3d"]["point3"])),
                    "absolutePoint1": dict(
                        zip("xyz", label["angle3d"]["computedpoint1world"])
                    ),
                    "absoluteVertex": dict(
                        zip("xyz", label["angle3d"]["computedpoint2world"])
                    ),
                    "absolutePoint2": dict(
                        zip("xyz", label["angle3d"]["computedpoint3world"])
                    ),
                    "normal": label["angle3d"]["normal"],
                    "angle": label["angle3d"]["computedangledeg"],
                    **label_obj,  # type: ignore
                }
            )
        elif label.get("point"):
            volume["landmarks"] = volume.get("landmarks", [])
            volume["landmarks"].append(
                {
                    "point": {
                        "xNorm": label["point"]["xnorm"],
                        "yNorm": label["point"]["ynorm"],
                    },
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                }
            )
        elif label.get("point3d"):
            volume["landmarks3d"] = volume.get("landmarks3d", [])
            volume["landmarks3d"].append(
                {
                    "point": {
                        "i": label["point3d"]["pointx"],
                        "j": label["point3d"]["pointy"],
                        "k": label["point3d"]["pointz"],
                    },
                    **label_obj,  # type: ignore
                }
            )
        elif label.get("polyline"):
            volume["polylines"] = volume.get("polylines", [])
            volume["polylines"].append(
                {
                    "points": [
                        {"xNorm": point["xnorm"], "yNorm": point["ynorm"]}
                        for point in label["polyline"]
                    ],
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                }
            )
        elif label.get("ellipse"):
            volume["ellipses"] = volume.get("ellipses", [])
            volume["ellipses"].append(
                {
                    "pointCenter": {
                        "xNorm": label["ellipse"]["xcenternorm"],
                        "yNorm": label["ellipse"]["ycenternorm"],
                    },
                    "xRadiusNorm": label["ellipse"]["xnorm"],
                    "yRadiusNorm": label["ellipse"]["ynorm"],
                    "rotationRad": label["ellipse"]["rot"],
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                    **measurement_stats,
                }
            )
        elif label.get("bbox2d"):
            volume["boundingBoxes"] = volume.get("boundingBoxes", [])
            volume["boundingBoxes"].append(
                {
                    "pointTopLeft": {
                        "xNorm": label["bbox2d"]["xnorm"],
                        "yNorm": label["bbox2d"]["ynorm"],
                    },
                    "wNorm": label["bbox2d"]["wnorm"],
                    "hNorm": label["bbox2d"]["hnorm"],
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                    **measurement_stats,
                }
            )
        elif label.get("cuboid"):
            volume["cuboids"] = volume.get("cuboids", [])
            volume["cuboids"].append(
                {
                    "point1": dict(zip("ijk", label["cuboid"]["point1"])),
                    "point2": dict(zip("ijk", label["cuboid"]["point2"])),
                    "absolutePoint1": dict(
                        zip("xyz", label["cuboid"]["computedpoint1world"])
                    ),
                    "absolutePoint2": dict(
                        zip("xyz", label["cuboid"]["computedpoint2world"])
                    ),
                    **label_obj,  # type: ignore
                    **measurement_stats,  # type: ignore
                }
            )
        elif label.get("polygon"):
            volume["polygons"] = volume.get("polygons", [])
            volume["polygons"].append(
                {
                    "points": [
                        {"xNorm": point["xnorm"], "yNorm": point["ynorm"]}
                        for point in label["polygon"]
                    ],
                    **label_obj,  # type: ignore
                    **video_metadata,  # type: ignore
                    **measurement_stats,  # type: ignore
                }
            )


def dicom_rb_format(
    task: Dict,
    taxonomy: Dict,
    old_format: bool,
    no_consensus: bool,
    review_stages: List[ReviewStage],
) -> TaskType.OutputTask:
    """Get new dicom rb task format."""
    # pylint: disable=too-many-branches, too-many-statements, too-many-locals, unused-argument
    if old_format:
        keys = ["itemsPresigned", "seriesInfo", "storageId", "labelStorageId"]
        for key in [
            "currentStageName",
            "priority",
            "createdBy",
            "createdAt",
            "updatedBy",
            "updatedAt",
            "labels",
            "labelsMap",
            "metaData",
        ]:
            if not task.get(key):
                keys.append(key)
        return {key: value for key, value in task.items() if key not in keys}  # type: ignore

    if not task.get("seriesInfo"):
        task["seriesInfo"] = [{"itemsIndices": list(range(len(task["items"])))}]

    output: TaskType.OutputTask = {"taskId": task["taskId"], "name": "", "series": []}

    if task.get("name"):
        output["name"] = task["name"]
    elif task.get("items"):
        output["name"] = task["items"][0]

    if task.get("currentStageName"):
        output["currentStageName"] = task["currentStageName"]

    if task.get("priority") is not None:
        output["priority"] = task["priority"]

    if task.get("createdBy"):
        output["createdBy"] = task["createdBy"]

    if task.get("createdAt"):
        output["createdAt"] = task["createdAt"]

    if task.get("updatedBy"):
        output["updatedBy"] = task["updatedBy"]

    if task.get("updatedAt"):
        output["updatedAt"] = task["updatedAt"]

    if task.get("metaData"):
        output["metaData"] = task["metaData"]

    volume_series: List[TaskType.Series] = [{} for _ in range(len(task["seriesInfo"]))]
    item_index_map: Dict[int, int] = {}
    for volume_index, series_info in enumerate(task["seriesInfo"]):
        series = volume_series[volume_index]
        if series_info.get("name"):
            series["name"] = series_info["name"]  # type: ignore

        series_meta_data = series_info.get("metaData")
        if isinstance(series_meta_data, str):
            series["metaData"] = json.loads(series_meta_data)

        series["items"] = []
        for item_index in series_info["itemsIndices"]:
            item_index_map[item_index] = volume_index
            series["items"].append(task["items"][item_index])  # type: ignore

    output["series"] = deepcopy(volume_series)
    if no_consensus:
        if task.get("consensusTasks"):
            consensus_task = task["consensusTasks"][0]
            task["labels"] = consensus_task.get("labels")
            task["labelsMap"] = consensus_task.get("labelsMap")
        dicom_rb_series(item_index_map, task, output, taxonomy)
    else:
        output["consensus"] = True
        if "consensusScore" in task:  # Review_1, END
            output["consensusScore"] = task["consensusScore"]

        if (
            task["currentStageName"] in [stage.stage_name for stage in review_stages]
            and "consensusScore" not in task
        ) or task[
            "currentStageName"
        ] == "END":  # Review_2...n, END
            super_truth: TaskType.OutputTask = {  # type: ignore
                "series": [{**series} for series in volume_series]
            }
            output["superTruth"] = super_truth
            if task.get("updatedBy"):
                output["superTruth"]["updatedBy"] = task["updatedBy"]
            if task.get("updatedAt"):
                output["superTruth"]["updatedAt"] = task["updatedAt"]
            dicom_rb_series(item_index_map, task, output["superTruth"], taxonomy)

        consensus_tasks: List[TaskType.OutputTask] = [
            {"series": [{**series} for series in volume_series]}  # type: ignore
            for _ in range(len(task.get("consensusTasks", []) or []))
        ]
        output["consensusTasks"] = consensus_tasks
        for consensus_idx, output_consensus_task in enumerate(output["consensusTasks"]):
            consensus_task = task["consensusTasks"][consensus_idx]
            if consensus_task.get("status"):
                output_consensus_task["status"] = consensus_task["status"]

            if consensus_task.get("assignee"):
                output_consensus_task["updatedBy"] = consensus_task["assignee"]
            elif consensus_task.get("email"):
                output_consensus_task["updatedBy"] = consensus_task["email"]

            if consensus_task.get("userId"):
                output_consensus_task["updatedByUserId"] = consensus_task["userId"]
            if consensus_task.get("updatedAt"):
                output_consensus_task["updatedAt"] = consensus_task["updatedAt"]

            if consensus_task.get("scores"):
                output_consensus_task["scores"] = []
                for consensus_score in consensus_task["scores"] or []:
                    score: TaskType.ConsensusScore = {}
                    if consensus_score.get("userId"):
                        score["secondaryUserId"] = consensus_score["userId"]

                    if consensus_score.get("assignee"):
                        score["secondaryUser"] = consensus_score["assignee"]
                    elif consensus_score.get("email"):
                        score["secondaryUserEmail"] = consensus_score["email"]

                    if consensus_score.get("score") is not None:
                        score["score"] = consensus_score["score"]
                    output_consensus_task["scores"].append(score)

            dicom_rb_series(
                item_index_map, consensus_task, output_consensus_task, taxonomy
            )

    return output



"""Graphql Client responsible for make API requests."""

import time
import json
import base64
import gzip
from typing import Dict
import requests  # type: ignore

import aiohttp
import tenacity
from tenacity.retry import retry_if_not_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_exponential

from redbrick import __version__ as sdk_version  # pylint: disable=cyclic-import
from redbrick.utils.logging import assert_validation, log_error, logger
from redbrick.common.constants import (
    DEFAULT_URL,
    MAX_RETRY_ATTEMPTS,
    REQUEST_TIMEOUT,
    PEERLESS_ERRORS,
)


class RBClient:
    """Client to communicate with RedBrick AI GraphQL Server."""

    def __init__(self, api_key: str, url: str) -> None:
        """Construct RBClient."""
        self.url = (url or DEFAULT_URL).lower().rstrip("/")
        if DEFAULT_URL in self.url:
            self.url = DEFAULT_URL
        elif "amazonaws.com" not in self.url and "localhost" not in self.url:
            self.url = self.url.replace("https://", "", 1).replace("http://", "", 1)
            pos = self.url.find("/")
            pos = pos if pos >= 0 else len(self.url)
            self.url = "https://" + self.url[:pos] + "/api"

        self.url += "/graphql/"
        self.session = requests.Session()

        self.api_key = api_key
        assert_validation(
            len(self.api_key) == 43,
            "Invalid Api Key length, make sure you've copied it correctly",
        )

    def __del__(self) -> None:
        """Garbage collect and close session."""
        self.session.close()

    @property
    def headers(self) -> Dict:
        """Get request headers."""
        return {
            "RB-SDK-Version": sdk_version,
            "ApiKey": self.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Content-Encoding-RB": "gzip",
            "Accept-Encoding": "br, gzip",
        }

    def prepare_query(self, query: str, variables: Dict) -> bytes:
        """Prepare query to be sent to the server."""
        return base64.b64encode(
            gzip.compress(
                json.dumps(
                    {"query": query, "variables": variables}, separators=(",", ":")
                ).encode("utf-8")
            )
        )

    @tenacity.retry(
        reraise=True,
        stop=stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_not_exception_type(PEERLESS_ERRORS),
    )
    def execute_query(
        self, query: str, variables: Dict, raise_for_error: bool = True
    ) -> Dict:
        """Execute a graphql query."""
        start_time = time.time()
        logger.debug("Executing: " + query.strip().split("\n")[0])
        response = self.session.post(
            self.url,
            timeout=REQUEST_TIMEOUT,
            headers=self.headers,
            data=self.prepare_query(query, variables),
        )
        self._check_status_msg(response.status_code, start_time)
        return self._process_json_response(response.json(), raise_for_error)

    @tenacity.retry(
        reraise=True,
        stop=stop_after_attempt(MAX_RETRY_ATTEMPTS),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_not_exception_type(PEERLESS_ERRORS),
    )
    async def execute_query_async(
        self,
        aio_session: aiohttp.ClientSession,
        query: str,
        variables: Dict,
        raise_for_error: bool = True,
    ) -> Dict:
        """Execute a graphql query using asyncio."""
        start_time = time.time()
        logger.debug("Executing async: " + query.strip().split("\n")[0])
        async with aio_session.post(
            self.url,
            timeout=REQUEST_TIMEOUT,
            headers=self.headers,
            data=self.prepare_query(query, variables),
        ) as response:
            self._check_status_msg(response.status, start_time)
            return self._process_json_response(await response.json(), raise_for_error)

    @staticmethod
    def _check_status_msg(response_status: int, start_time: float) -> None:
        total_time = time.time() - start_time
        logger.debug(f"Response status: {response_status} took {total_time} seconds")
        if response_status == 413 or response_status >= 500:
            if response_status == 413 or total_time >= 26:
                raise TimeoutError(
                    "Request timed out/too large. Please consider using lower concurrency"
                )
            raise ConnectionError(
                "Internal Server Error: You are probably using an invalid API key"
            )
        if response_status == 403:
            raise PermissionError("Problem authenticating with Api Key")

    @staticmethod
    def _process_json_response(
        response_data: Dict, raise_for_error: bool = True
    ) -> Dict:
        """Process JSON resonse."""
        if "errors" in response_data:
            errors = []
            for error in response_data["errors"]:
                errors.append(error["message"])
                log_error(error["message"])

            if raise_for_error:
                raise ValueError("\n".join(errors))

            del response_data["errors"]

        res = {}
        if "data" in response_data:
            res = response_data["data"]
        else:
            res = response_data
        return res



"""CLI upload command."""

import os
import re
import json
import asyncio
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import List, Dict, Optional, Union, cast

from redbrick.cli.input.select import CLIInputSelect
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIUploadInterface
from redbrick.common.enums import StorageMethod, ImportTypes
from redbrick.utils.logging import assert_validation, logger
from redbrick.utils.files import find_files_recursive
from redbrick.types.task import InputTask


class CLIUploadController(CLIUploadInterface):
    """CLI upload command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize upload sub commands."""
        parser.add_argument(
            "directory",
            help="The directory containing files to upload to the project",
        )
        parser.add_argument(
            "--as-frames",
            action="store_true",
            help="Upload video from image frames",
        )
        parser.add_argument(
            "--type",
            "-t",
            nargs="?",
            default=ImportTypes.DICOM3D.value,
            choices=[import_type.value for import_type in ImportTypes],
            help=f"""Import file type
{['`' + import_type.value + '`' for import_type in ImportTypes]}\n
Please refer to [our documentation](https://docs.redbrickai.com/importing-data/direct-data-upload),
to understand the required folder structure and supported file types.
""",
        )
        parser.add_argument(
            "--as-study",
            action="store_true",
            help="Group files by study",
        )
        parser.add_argument(
            "--json",
            action="store_true",
            help="Upload json files with list of task objects",
        )
        parser.add_argument(
            "--segment-map",
            "-m",
            help="Segmentation mapping file path",
        )
        parser.add_argument(
            "--storage",
            "-s",
            default=self.STORAGE_REDBRICK,
            help="Storage method: "
            + f"({self.STORAGE_REDBRICK} [default], {self.STORAGE_PUBLIC}, <storage id>)",
        )
        parser.add_argument(
            "--label-storage",
            help="Label Storage method: (same as items storage `--storage` [default],"
            + f" {self.STORAGE_REDBRICK}, {self.STORAGE_PUBLIC}, <storage id>)",
        )
        parser.add_argument(
            "--ground-truth",
            action="store_true",
            help="""Upload tasks directly to ground truth.""",
        )
        parser.add_argument(
            "--label-validate",
            action="store_true",
            help="""
Validate NIfTI label instances and segmentMap.
By default, the uploaded NIfTI files are not validated during upload,
which can result in invalid files being uploaded.
Using this argument validates the files before upload,
but may increase the upload time.""",
        )
        parser.add_argument(
            "--rt-struct",
            action="store_true",
            help="Upload segmentations from DICOM RT-Struct files.",
        )
        parser.add_argument(
            "--clear-cache", action="store_true", help="Clear local cache"
        )
        parser.add_argument(
            "--concurrency",
            "-c",
            type=int,
            default=10,
            help="Concurrency value (Default: 10)",
        )

    def handler(self, args: Namespace) -> None:
        """Handle upload command."""
        self.args = args
        project = CLIProject.from_path()
        assert_validation(project, "Not a valid project")
        self.project = cast(CLIProject, project)

        self.handle_upload()

    def handle_upload(self) -> None:  # noqa: ignore=C901
        """Handle empty sub command."""
        # pylint: disable=protected-access, too-many-branches, too-many-locals, too-many-statements
        # pylint: disable=too-many-nested-blocks
        logger.debug("Uploading data to project")
        project = self.project.project

        directory = os.path.normpath(self.args.directory)
        if directory.endswith(".json") and os.path.isfile(directory):
            self.args.json = True
        elif not os.path.isdir(directory):
            raise Exception(f"{directory} is not a valid directory")

        if self.args.storage == self.STORAGE_REDBRICK:
            storage_id = str(StorageMethod.REDBRICK)
        elif self.args.storage == self.STORAGE_PUBLIC:
            storage_id = str(StorageMethod.PUBLIC)
        elif re.match(
            r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$",
            self.args.storage.strip().lower(),
        ):
            storage_id = str(self.args.storage).strip().lower()
        else:
            raise ArgumentError(None, f"Invalid upload storage: {self.args.storage}")

        if not self.args.label_storage:
            label_storage_id = storage_id
        elif self.args.label_storage == self.STORAGE_REDBRICK:
            label_storage_id = str(StorageMethod.REDBRICK)
        elif self.args.label_storage == self.STORAGE_PUBLIC:
            label_storage_id = str(StorageMethod.PUBLIC)
        elif re.match(
            r"^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$",
            self.args.label_storage.strip().lower(),
        ):
            label_storage_id = str(self.args.label_storage).strip().lower()
        else:
            raise ArgumentError(
                None, f"Invalid upload label storage: {self.args.label_storage}"
            )

        if self.args.clear_cache:
            self.project.cache.clear_cache(True)

        items_list: Union[List[List[str]], List[Dict]]
        if self.args.json and directory.endswith(".json") and os.path.isfile(directory):
            items_list = [[os.path.abspath(directory)]]
        else:
            logger.info(f"Searching for items recursively in {directory}")
            items_list = find_files_recursive(
                directory, set(["json" if self.args.json else "*"])
            )

        logger.debug(f"Contains {len(items_list)} items")

        upload_cache_hash = self.project.conf.get_option("uploads", "cache")
        upload_cache = set(
            self.project.cache.get_data("uploads", upload_cache_hash, True, True) or []
        )

        loop = asyncio.get_event_loop()

        if not self.args.json and items_list:
            import_file_type = CLIInputSelect(
                self.args.type,
                "Import file type",
                [import_type.value for import_type in ImportTypes],
            ).get()

            items_list = loop.run_until_complete(
                self.project.project.upload.generate_items_list(
                    items_list,
                    import_file_type,
                    self.args.as_study,
                    self.args.concurrency,
                )
            )

        segmentation_mapping = {}
        if self.args.segment_map:
            with open(self.args.segment_map, "r", encoding="utf-8") as file_:
                segmentation_mapping = json.load(file_)
                assert_validation(
                    isinstance(segmentation_mapping, dict),
                    "Segmentation mapping is invalid",
                )

        points: List[Dict] = []
        uploading = set()
        if self.args.json:
            files_data: List[List[InputTask]] = []
            task_dirs: List[str] = []
            for item_group in items_list:
                task_dirs.append(os.path.dirname(item_group[0]))
                with open(item_group[0], "r", encoding="utf-8") as file_:
                    files_data.append(json.load(file_))
            logger.debug("Preparing json files for upload")
            points = self.project.project.upload.prepare_json_files(
                files_data,
                storage_id,
                label_storage_id,
                segmentation_mapping,
                task_dirs,
                upload_cache,
                self.args.rt_struct,
                self.args.label_validate,
                self.args.concurrency,
            )
            segmentation_mapping = {}
        else:
            for igroup in items_list:
                if isinstance(igroup, dict):
                    item_name = igroup.get("name")
                    item_group = list(
                        map(os.path.normpath, igroup.get("items", []) or [])
                    )
                else:
                    item_name = None
                    item_group = igroup

                if not item_group:
                    continue

                if len(item_group) > 1:
                    items_dir = os.path.dirname(item_group[0])
                    task_dir = os.path.dirname(items_dir)
                    task_name = os.path.basename(items_dir)
                else:
                    task_dir = os.path.dirname(item_group[0])
                    task_name = os.path.splitext(os.path.basename(item_group[0]))[0]

                label_data: Optional[Dict] = None

                label_file = os.path.join(task_dir, task_name + ".json")
                if os.path.isfile(label_file):
                    with open(label_file, "r", encoding="utf-8") as file_:
                        label_data = json.load(file_)

                item_name = (
                    label_data["name"]
                    if label_data and label_data.get("name")
                    else (
                        item_name
                        if item_name
                        else (
                            re.sub(
                                r"^" + re.escape(directory + os.path.sep) + r"?",
                                "",
                                (
                                    os.path.dirname(item_group[0])
                                    if len(item_group) > 1
                                    else item_group[0]
                                ),
                            )
                            or directory
                        ).replace(os.path.sep, "/")
                    )
                )
                if item_name in upload_cache or item_name in uploading:
                    logger.info(f"Skipping duplicate item name: {item_name}")
                    continue

                uploading.add(item_name)

                item = {
                    "name": item_name,
                    "items": item_group[:],
                    "labels": label_data.get("labels", []) if label_data else [],
                }
                if label_data and label_data.get("segmentMap"):
                    item["segmentMap"] = label_data["segmentMap"]

                if label_data and isinstance(label_data, dict):
                    if (
                        "segmentations" in label_data
                        and "labelsMap" not in label_data
                        and isinstance(label_data["segmentations"], (list, dict))
                    ):
                        if isinstance(label_data["segmentations"], list):
                            label_data["segmentations"] = {
                                str(idx): segmentation
                                for idx, segmentation in enumerate(
                                    label_data["segmentations"]
                                )
                            }
                        if isinstance(label_data["segmentations"], dict):
                            label_data["labelsMap"] = [
                                (
                                    {"labelName": segmentation, "imageIndex": int(idx)}
                                    if segmentation
                                    else None
                                )
                                for idx, segmentation in label_data[
                                    "segmentations"
                                ].items()
                            ]
                    elif "labelsPath" in label_data and "labelsMap" not in label_data:
                        label_data["labelsMap"] = [
                            {
                                "labelName": label_data["labelsPath"],
                                "imageIndex": 0,
                            }
                        ]
                    if label_data.get("labelsMap"):
                        item["labelsMap"] = []
                        for label_map in label_data["labelsMap"]:
                            if not isinstance(label_map, dict):
                                label_map = {}
                            if not isinstance(label_map["labelName"], list):
                                label_map["labelName"] = [label_map["labelName"]]
                            label_map["labelName"] = [
                                (
                                    label_name
                                    if os.path.isabs(label_name)
                                    or not os.path.exists(
                                        os.path.join(task_dir, label_name)
                                    )
                                    else os.path.abspath(
                                        os.path.join(task_dir, label_name)
                                    )
                                )
                                for label_name in label_map["labelName"]
                            ]
                            if len(label_map["labelName"]) == 1:
                                label_map["labelName"] = label_map["labelName"][0]
                            item["labelsMap"].append(
                                {
                                    "labelName": label_map["labelName"],
                                    "imageIndex": int(label_map["imageIndex"]),
                                    "imageName": label_map.get("imageName"),
                                    "seriesId": label_map.get("seriesId"),
                                }
                            )

                points.append(item)

        if points:
            logger.info(f"Found {len(points)} items")

            uploads = loop.run_until_complete(
                project.upload._create_tasks(
                    points,
                    segmentation_mapping,
                    self.args.ground_truth,
                    storage_id,
                    label_storage_id,
                    self.args.label_validate,
                    self.args.concurrency,
                    False,
                    self.args.rt_struct,
                )
            )

            for upload in uploads:
                if upload.get("response"):
                    upload_cache.add(upload["name"])

            self.project.conf.set_option(
                "uploads",
                "cache",
                self.project.cache.set_data("uploads", list(upload_cache), True),
            )
            self.project.conf.save()
        elif not items_list:
            logger.info(
                "No items found. Please ensure that you have specified the correct data type: "
                + "DICOM3D, NIFTI3D, etc. Type `redbrick upload -h` for more options and usage"
            )



"""Repo for accessing export apis."""

from typing import Any, Optional, List, Dict, Sequence, Tuple
from datetime import datetime
from dateutil import parser  # type: ignore

from redbrick.common.export import ExportControllerInterface, TaskFilterParams
from redbrick.common.client import RBClient
from redbrick.repo.shards import datapoint_shard, task_shard, router_task_shard


class ExportRepo(ExportControllerInterface):
    """Handle API requests to get export data."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def datapoints_in_project(
        self, org_id: str, project_id: str, stage_name: Optional[str] = None
    ) -> int:
        """Get number of datapoints in project."""
        query_string = """
        query tasksPagedSDK($orgId: UUID!, $projectId: UUID!, $stageName: String, $first: Int) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                first: $first
            ) {
                count
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "first": 0,
        }

        result = self.client.execute_query(query_string, query_variables)

        return int(result["tasksPaged"]["count"])

    def get_datapoint_latest(
        self,
        org_id: str,
        project_id: str,
        task_id: str,
        presign_items: bool = False,
        with_consensus: bool = False,
    ) -> Dict:
        """Get the latest datapoint."""
        query_string = f"""
        query taskSDK($orgId: UUID!, $projectId: UUID!, $taskId: UUID!) {{
            task(orgId: $orgId, projectId: $projectId, taskId: $taskId) {{
                {task_shard(presign_items, with_consensus)}
            }}
        }}
        """
        # EXECUTE THE QUERY
        query_variables = {"orgId": org_id, "projectId": project_id, "taskId": task_id}

        result = self.client.execute_query(query_string, query_variables)

        return result["task"]

    def get_datapoints_latest(
        self,
        org_id: str,
        project_id: str,
        stage_name: Optional[str] = None,
        cache_time: Optional[datetime] = None,
        presign_items: bool = False,
        with_consensus: bool = False,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str], Optional[datetime]]:
        """Get the latest datapoints."""
        # pylint: disable=too-many-locals
        query_string = f"""
        query tasksPagedSDK(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String
            $cacheTime: DateTime
            $first: Int
            $after: String
        ) {{
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                cacheTime: $cacheTime
                first: $first
                after: $after
            ) {{
                entries {{
                    {task_shard(presign_items, with_consensus)}
                }}
                cursor
                cacheTime
            }}
        }}
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cacheTime": None if cache_time is None else cache_time.isoformat(),
            "first": first,
            "after": cursor,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        tasks_paged = result.get("tasksPaged", {}) or {}
        entries: List[Dict] = tasks_paged.get("entries", []) or []  # type: ignore
        new_cache_time: Optional[str] = tasks_paged.get("cacheTime")
        return (
            entries,
            tasks_paged.get("cursor"),
            parser.parse(new_cache_time) if new_cache_time else None,
        )

    def task_search(
        self,
        org_id: str,
        project_id: str,
        stage_name: Optional[str] = None,
        task_search: Optional[str] = None,
        manual_labeling_filters: Optional[TaskFilterParams] = None,
        only_meta_data: bool = True,
        first: int = 50,
        after: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Task search."""
        query_string = f"""
        query tasksListSDK(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String
            $taskSearch: String
            $manualLabelingFilters: TasksFilter
            $first: Int
            $after: String
        ) {{
            genericTasks(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskSearch: $taskSearch
                manualLabelingFilters: $manualLabelingFilters
                first: $first
                after: $after
            ) {{
                entries {{
                    taskId
                    currentStageName
                    createdAt
                    updatedAt
                    priority
                    datapoint {{
                        {datapoint_shard(not only_meta_data, not only_meta_data)}
                    }}
                    currentStageSubTask {{
                        ... on LabelingTask {{
                            assignedTo {{
                                userId
                            }}
                            state
                            assignedAt
                            progressSavedAt
                            completedAt
                            completionTimeMs
                            subTasks {{
                                assignedTo {{
                                    userId
                                }}
                                state
                                assignedAt
                                progressSavedAt
                                completedAt
                                completionTimeMs
                            }}
                        }}
                    }}
                }}
                cursor
            }}
        }}
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskSearch": task_search,
            "manualLabelingFilters": manual_labeling_filters,
            "first": first,
            "after": after,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        generic_tasks = result.get("genericTasks", {}) or {}
        entries: List[Dict] = generic_tasks.get("entries", []) or []  # type: ignore

        return entries, generic_tasks.get("cursor")

    def presign_items(
        self, org_id: str, storage_id: str, items: Sequence[Optional[str]]
    ) -> List[Optional[str]]:
        """Presign download items."""
        query = """
        query presignItemsSDK(
            $orgId: UUID!
            $storageId: UUID!
            $items: [String]!
        ) {
            presignItems(orgId: $orgId, storageId: $storageId, items: $items)
        }
        """

        variables = {"orgId": org_id, "storageId": storage_id, "items": items}

        response = self.client.execute_query(query, variables)
        presigned_items: List[Optional[str]] = response.get("presignItems", [])
        return presigned_items

    def task_events(
        self,
        org_id: str,
        project_id: str,
        task_id: Optional[str] = None,
        stage_name: Optional[str] = None,
        cache_time: Optional[datetime] = None,
        with_labels: bool = False,
        first: int = 10,
        after: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get task events."""
        query_variables: Dict[str, Any]
        if task_id:
            query_string = f"""
            query taskSDK($orgId: UUID!, $projectId: UUID!, $taskId: UUID!) {{
                task(orgId: $orgId, projectId: $projectId, taskId: $taskId) {{
                    {router_task_shard(with_labels)}
                }}
            }}
            """
            query_variables = {
                "orgId": org_id,
                "projectId": project_id,
                "taskId": task_id,
            }

            result = self.client.execute_query(query_string, query_variables, True)
            task = result.get("task") or {}
            return [task] if task else [], None

        query_string = f"""
        query taskEventsSDK(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String
            $cacheTime: DateTime
            $first: Int
            $after: String
        ) {{
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                cacheTime: $cacheTime
                first: $first
                after: $after
            ) {{
                entries {{
                    {router_task_shard(with_labels)}
                }}
                cursor
            }}
        }}
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cacheTime": None if cache_time is None else cache_time.isoformat(),
            "first": first,
            "after": after,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        task_events = result.get("tasksPaged") or {}
        entries: List[Dict] = task_events.get("entries") or []
        return entries, task_events.get("cursor")

    def active_time(
        self,
        org_id: str,
        project_id: str,
        stage_name: str,
        task_id: Optional[str] = None,
        first: int = 100,
        after: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get task active time."""
        query_string = """
        query taskActiveTimeSDK(
            $orgId: UUID!
            $projectId: UUID!
            $stageName: String!
            $taskId: UUID
            $first: Int
            $after: String
        ) {
            taskActiveTime(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                taskId: $taskId
                first: $first
                after: $after
            ) {
                entries {
                    taskId
                    user {
                        userId
                    }
                    timeSpent
                    cycle
                    date
                }
                cursor
            }
        }
        """

        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "taskId": task_id,
            "first": first,
            "after": after,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        task_times = result.get("taskActiveTime", {}) or {}
        entries: List[Dict] = task_times.get("entries", []) or []
        return entries, task_times.get("cursor")



"""Async utils."""

import asyncio
from typing import Any, Awaitable, Coroutine, List, Tuple, TypeVar, Optional, Iterable
import tqdm.asyncio  # type: ignore

from redbrick.common.constants import MAX_CONCURRENCY

ReturnType = TypeVar("ReturnType")  # pylint: disable=invalid-name


async def gather_with_concurrency(
    max_concurrency: int,
    tasks: Iterable[Awaitable[ReturnType]],
    progress_bar_name: Optional[str] = None,
    keep_progress_bar: bool = True,
    return_exceptions: bool = False,
) -> List[ReturnType]:
    """Utilizes a Semaphore to limit concurrency to n."""
    # pylint: disable=too-many-locals
    if not tasks:
        return []

    max_concurrency = max(1, min(max_concurrency, MAX_CONCURRENCY))

    if max_concurrency == 1:
        output: List[ReturnType] = []
        progress = (
            tqdm.tqdm(tasks, desc=progress_bar_name, leave=keep_progress_bar)
            if progress_bar_name
            else tasks
        )
        for task in progress:
            try:
                output.append(await task)
            except Exception as exc:  # pylint: disable=broad-except
                if return_exceptions:
                    output.append(exc)  # type: ignore
                else:
                    raise exc
        return output

    semaphore = asyncio.Semaphore(max_concurrency)

    async def sem_task(task: Awaitable[ReturnType]) -> ReturnType:
        async with semaphore:
            return await task

    coros = [sem_task(task) for task in tasks]
    if not progress_bar_name:
        return await asyncio.gather(  # type: ignore
            *coros, return_exceptions=return_exceptions
        )

    async def ordered_coroutine(
        idx: int, task: Coroutine[Any, Any, ReturnType]
    ) -> Tuple[int, bool, ReturnType]:
        """Return the index and result of a task."""
        try:
            return idx, True, await task
        except Exception as exc:  # pylint: disable=broad-except
            return idx, False, exc  # type: ignore

    ordered_coros = [ordered_coroutine(idx, task) for idx, task in enumerate(coros)]
    result: List[Tuple[int, ReturnType]] = []

    for coro in tqdm.asyncio.tqdm.as_completed(
        ordered_coros, desc=progress_bar_name, leave=keep_progress_bar
    ):
        temp: Coroutine[Any, Any, Tuple[int, bool, ReturnType]] = coro
        idx, success, value = await temp
        if not success and not return_exceptions:
            raise value  # type: ignore

        result.append((idx, value))

    return [res[1] for res in sorted(result, key=lambda x: x[0])]



"""Dicom/nifti related functions."""

import os
from typing import Any, Dict, List, Optional, Set, Tuple, Union, TypedDict
from asyncio import BoundedSemaphore
import shutil
from uuid import uuid4

from redbrick.utils.common_utils import config_path
from redbrick.utils.files import uniquify_path
from redbrick.utils.logging import log_error, logger
from redbrick.types.task import SegmentMap as TypeSegmentMap

# pylint: disable=too-many-locals, import-outside-toplevel, too-many-branches, too-many-statements


semaphore = BoundedSemaphore(1)


class LabelMapData(TypedDict):
    """Label map data."""

    semantic_mask: bool
    binary_mask: bool
    png_mask: bool
    masks: Optional[Union[str, List[str]]]


def merge_segmentations(
    input_file: str,
    input_instance: int,
    equals: bool,
    output_file: str,
    output_instance: int,
) -> bool:
    """Merge segmentations from input to output."""
    import numpy  # type: ignore
    from nibabel.loadsave import load as nib_load, save as nib_save  # type: ignore
    from nibabel.nifti1 import Nifti1Image  # type: ignore

    try:
        input_img = nib_load(input_file)

        if not isinstance(input_img, Nifti1Image):
            log_error(f"{input_file} is not a valid NIfTI1 file.")
            return False

        input_data = input_img.get_fdata(caching="unchanged")

        if os.path.isfile(output_file):
            output_img = nib_load(output_file)

            if not isinstance(output_img, Nifti1Image):
                log_error(f"{output_img} is not a valid NIfTI1 file.")
                return False

            output_affine = output_img.affine
            output_header = output_img.header
            output_data = output_img.get_fdata(caching="unchanged")
            output_data = numpy.round(output_data).astype(numpy.uint16)
        else:
            output_affine = input_img.affine
            output_header = input_img.header
            output_data = numpy.zeros(input_data.shape, dtype=numpy.uint16).astype(
                numpy.uint16
            )

        if equals:
            output_data[input_data == input_instance] = output_instance
        else:
            output_data[input_data != input_instance] = output_instance
        new_img = Nifti1Image(output_data, output_affine, output_header)
        new_img.set_data_dtype(output_data.dtype)
        nib_save(new_img, output_file)
        return True
    except Exception as err:  # pylint: disable=broad-except
        log_error(err)
        return False


def convert_to_binary(
    mask: str, labels: List[Dict], dirname: str
) -> Tuple[bool, List[str]]:
    """Convert segmentation to binary."""
    import numpy  # type: ignore
    from nibabel.loadsave import load as nib_load, save as nib_save  # type: ignore
    from nibabel.nifti1 import Nifti1Image  # type: ignore

    img = nib_load(mask)

    if not isinstance(img, Nifti1Image):
        log_error(f"{mask} is not a valid NIfTI1 file.")
        return False, [mask]

    affine = img.affine
    header = img.header
    data = img.get_fdata(caching="unchanged")

    files: List[str] = []

    for label in labels:
        instances: Set[int] = set(
            [label["dicom"]["instanceid"]] + (label["dicom"].get("groupids", []) or [])
        )
        new_data = numpy.zeros(
            data.shape, dtype=numpy.uint8 if max(instances) <= 255 else numpy.uint16
        )
        for instance in instances:
            new_data[data == instance] = 1

        if not numpy.any(new_data):
            continue

        filename = os.path.join(
            dirname, f"instance-{label['dicom']['instanceid']}.nii.gz"
        )
        if os.path.isfile(filename):
            os.remove(filename)

        new_img = Nifti1Image(new_data, affine, header)
        if new_data.dtype == numpy.uint16:
            new_img.set_data_dtype(numpy.uint16)
        nib_save(new_img, filename)
        files.append(filename)

    return True, files


def convert_to_semantic(
    masks: List[str],
    taxonomy: Dict,
    labels: List[Dict],
    dirname: str,
    binary_mask: bool,
) -> Tuple[bool, List[str]]:
    """Convert segmentation to semantic."""
    if not taxonomy.get("isNew"):
        log_error("Taxonomy V1 is not supported")
        return False, masks

    if not (binary_mask or len(masks) == 1):
        log_error(f"Cannot process labels: {labels}")
        return False, masks

    input_filename = output_filename = masks[0]
    if not binary_mask:
        input_filename = f"{output_filename}.old.nii.gz"
        os.rename(output_filename, input_filename)

    visited: Set[int] = set()
    files: Set[str] = set()
    for label in labels:
        instance = label["dicom"]["instanceid"]
        category = label["classid"] + 1
        if binary_mask:
            input_filename = os.path.join(dirname, f"instance-{instance}.nii.gz")
            output_filename = os.path.join(dirname, f"category-{category}.nii.gz")
            merged = merge_segmentations(input_filename, 1, True, output_filename, 1)
        else:
            merged = merge_segmentations(
                input_filename, instance, True, output_filename, category
            )
            visited.add(instance)
            for group_id in label["dicom"].get("groupids", []) or []:
                if group_id in visited:
                    continue
                group_merged = merge_segmentations(
                    input_filename, group_id, True, output_filename, category
                )
                if group_merged:
                    visited.add(group_id)
                else:
                    log_error(f"Error processing semantic mask for: {label}")

        if merged:
            files.add(output_filename)
        else:
            log_error(f"Error processing semantic mask for: {label}")

    if not binary_mask:
        os.remove(input_filename)

    return True, list(files)


def convert_to_png(
    masks: List[str],
    color_map: Dict,
    labels: List[Dict],
    dirname: str,
    binary_mask: bool,
    semantic_mask: bool,
    is_tax_v2: bool,
) -> Tuple[bool, List[str]]:
    """Convert masks to png."""
    import numpy  # type: ignore
    from nibabel.loadsave import load as nib_load  # type: ignore
    from nibabel.nifti1 import Nifti1Image  # type: ignore
    from PIL import Image  # type: ignore

    cat_class_map: Dict[str, str] = {}
    instance_class_map: Dict[int, int] = {}
    for label in labels:
        instance_class_map[label["dicom"]["instanceid"]] = label["classid"] + 1
        category = "::".join(label["category"][0][1:])
        cat_class_map[f"category-{label['classid'] + 1}"] = category
        cat_class_map[f"instance-{label['dicom']['instanceid']}"] = category
        for group_id in label["dicom"].get("groupids", []) or []:
            if f"instance-{group_id}" in cat_class_map:
                continue
            cat_class_map[f"instance-{group_id}"] = category

    files: Set[str] = set()
    for mask in masks:
        mask_img = nib_load(mask)
        if not isinstance(mask_img, Nifti1Image):
            log_error(f"{mask} is not a valid NIfTI1 file.")
            continue

        input_filename = os.path.basename(mask)[:-7]
        mask_data = mask_img.get_fdata(caching="unchanged")
        if mask_data.shape[2] != 1:
            log_error(f"{mask} is not a 2D image")
            continue

        mask_arr = numpy.round(mask_data).astype(numpy.uint16).swapaxes(0, 1)
        mask_arr = mask_arr.reshape(mask_arr.shape[0], mask_arr.shape[1])
        if binary_mask:
            color_mask = numpy.zeros((mask_arr.shape[0], mask_arr.shape[1], 3))
            cat = int(input_filename.split("-")[-1])
            if not semantic_mask:
                cat = instance_class_map.get(cat, 0)
            color_mask[mask_arr == 1] = (
                color_map.get(cat - 1, (255, 255, 255))
                if is_tax_v2
                else color_map.get(
                    cat_class_map.get(input_filename, ""), (255, 255, 255)
                )
            )
            filename = os.path.join(dirname, f"mask-{cat}.png")
            pil_color_mask = Image.fromarray(color_mask.astype(numpy.uint8))
            pil_color_mask.save(filename)
            files.add(filename)

        else:
            color_mask = numpy.zeros((mask_arr.shape[0], mask_arr.shape[1], 3))
            for seg in numpy.unique(mask_arr):
                if not seg:
                    continue
                cat = int(seg)
                if not semantic_mask:
                    cat = instance_class_map.get(cat, 0)
                color_mask[mask_arr == seg] = (
                    color_map.get(cat - 1, (255, 255, 255))
                    if is_tax_v2
                    else color_map.get(
                        cat_class_map.get(input_filename, ""), (255, 255, 255)
                    )
                )
            filename = os.path.join(dirname, f"{input_filename}.png")
            pil_color_mask = Image.fromarray(color_mask.astype(numpy.uint8))
            pil_color_mask.save(filename)
            files.add(filename)

    return bool(files), list(files)


async def process_nifti_download(
    labels: List[Dict],
    labels_path: Optional[str],
    png_mask: bool,
    color_map: Dict,
    semantic_mask: bool,
    binary_mask: Optional[bool],
    taxonomy: Dict,
    volume_index: Optional[int],
) -> LabelMapData:
    """Process nifti download file."""
    label_map_data = LabelMapData(
        semantic_mask=False, binary_mask=False, png_mask=False, masks=labels_path
    )
    async with semaphore:
        try:
            if not (labels_path and os.path.isfile(labels_path)):
                return label_map_data

            filtered_labels = [
                label
                for label in labels
                if label.get("dicom")
                and (
                    volume_index is None
                    or label.get("volumeindex") is None
                    or label["volumeindex"] == volume_index
                )
            ]

            binary_mask = (
                binary_mask
                if binary_mask is not None
                else any(label["dicom"].get("groupids") for label in filtered_labels)
            )

            if not (png_mask or binary_mask or semantic_mask):
                return label_map_data

            dirname = (
                os.path.splitext(labels_path)[0]
                if labels_path.endswith(".gz")
                else labels_path
            )
            dirname = os.path.splitext(dirname)[0]
            shutil.rmtree(dirname, ignore_errors=True)
            os.makedirs(dirname, exist_ok=True)

            if binary_mask:
                (
                    label_map_data["binary_mask"],
                    label_map_data["masks"],
                ) = convert_to_binary(labels_path, filtered_labels, dirname)
            else:
                label_map_data["masks"] = [labels_path]

            if semantic_mask and label_map_data["masks"]:
                (
                    label_map_data["semantic_mask"],
                    label_map_data["masks"],
                ) = convert_to_semantic(
                    (
                        [label_map_data["masks"]]
                        if isinstance(label_map_data["masks"], str)
                        else label_map_data["masks"]
                    ),
                    taxonomy,
                    filtered_labels,
                    dirname,
                    label_map_data["binary_mask"],
                )

            if label_map_data["semantic_mask"]:
                for path in os.listdir(dirname):
                    if path.startswith("instance-"):
                        os.remove(os.path.join(dirname, path))

            if png_mask and label_map_data["masks"]:
                label_map_data["png_mask"], label_map_data["masks"] = convert_to_png(
                    (
                        [label_map_data["masks"]]
                        if isinstance(label_map_data["masks"], str)
                        else label_map_data["masks"]
                    ),
                    color_map,
                    filtered_labels,
                    dirname,
                    label_map_data["binary_mask"],
                    label_map_data["semantic_mask"],
                    bool(taxonomy.get("isNew")),
                )

            if label_map_data["png_mask"]:
                for path in os.listdir(dirname):
                    if path.startswith("instance-") or path.startswith("category-"):
                        os.remove(os.path.join(dirname, path))

            if not os.listdir(dirname):
                shutil.rmtree(dirname)

        except Exception as error:  # pylint: disable=broad-except
            log_error(f"Failed to process {labels_path}: {error}")

        return label_map_data


async def process_nifti_upload(
    files: Union[str, List[str]],
    instances: Set[int],
    binary_mask: bool,
    semantic_mask: bool,
    png_mask: bool,
    masks: Dict[str, str],
    label_validate: bool,
) -> Tuple[Optional[str], Dict[int, List[int]]]:
    """Process nifti upload files."""
    # pylint: disable=too-many-locals, too-many-branches, import-outside-toplevel
    # pylint: disable=too-many-statements, too-many-return-statements, unused-argument
    async with semaphore:
        import numpy  # type: ignore
        from nibabel.loadsave import load as nib_load, save as nib_save  # type: ignore
        from nibabel.nifti1 import Nifti1Image  # type: ignore
        from nibabel.nifti2 import Nifti2Image  # type: ignore

        if png_mask:
            log_error("PNG mask upload is not supported yet")
            return None, {}

        if isinstance(files, str):
            files = [files]
        if not files or any(
            not isinstance(file_, str) or not os.path.isfile(file_) for file_ in files
        ):
            return None, {}

        if len(files) == 1 and not label_validate:
            return files[0], {}

        try:
            reverse_masks: Dict[str, Tuple[int, ...]] = {}
            for inst_id, mask in masks.items():
                if mask not in reverse_masks:
                    reverse_masks[mask] = tuple()
                reverse_masks[mask] = reverse_masks[mask] + (int(inst_id),)

            delete_files: List[str] = []
            if binary_mask:
                for file_idx, file_path in enumerate(files):
                    new_filename = uniquify_path(f"{file_idx}.nii.gz")
                    if len(
                        reverse_masks.get(file_path, tuple()) or tuple()
                    ) == 1 and merge_segmentations(
                        file_path,
                        0,
                        False,
                        new_filename,
                        reverse_masks[file_path][0],
                    ):
                        delete_files.append(new_filename)
                        files[file_idx] = new_filename

            base_img = nib_load(files[0])
            instance_map: Dict[int, List[int]] = {}
            reverse_instance_map: Dict[Tuple[int, ...], int] = {}

            if not isinstance(base_img, Nifti1Image) and not isinstance(
                base_img, Nifti2Image
            ):
                return None, {}

            base_data = base_img.get_fdata(caching="unchanged")

            if base_img.get_data_dtype() != numpy.uint16:
                base_img.set_data_dtype(numpy.uint16)
                base_data = numpy.round(base_data).astype(numpy.uint16)  # type: ignore

            if base_data.ndim != 3:
                return None, {}

            unique_instances = numpy.unique(base_data)  # type: ignore
            for instance in unique_instances:
                inst = int(round(instance))
                if not inst:
                    continue
                if inst not in instance_map:
                    instance_map[inst] = [inst]
                    reverse_instance_map[(inst,)] = inst

            for img_idx, file_ in enumerate(files):
                if not img_idx:
                    continue

                img = nib_load(file_)
                data = img.get_fdata(caching="unchanged")  # type: ignore
                if base_data.shape != data.shape:
                    return None, {}

                unique_instances = numpy.unique(data)  # type: ignore
                for instance in unique_instances:
                    inst = int(round(instance))
                    if not inst:
                        continue
                    if inst not in instance_map:
                        instance_map[inst] = [inst]
                        reverse_instance_map[(inst,)] = inst

            used_instances = set(instance_map.keys())

            if label_validate and instances != used_instances:
                raise ValueError(
                    "Instance IDs in segmentation file(s) and segmentMap do not match.\n"
                    + f"Segmentation file(s) have instances: {used_instances} and "
                    + f"segmentMap has instances: {instances}\n"
                    + f"Segmentation(s): {files}"
                )

            group_instances = sorted(
                set(range(1, 65536)) - instances - used_instances, reverse=True
            )

            for file_ in files[1:]:
                img = nib_load(file_)
                if not isinstance(img, Nifti1Image) and not isinstance(
                    img, Nifti2Image
                ):
                    return None, {}

                data = img.get_fdata(caching="unchanged")
                if img.get_data_dtype() != numpy.uint16:
                    data = numpy.round(data).astype(numpy.uint16)  # type: ignore

                for i, j, k in zip(  # type: ignore
                    *numpy.where(numpy.logical_and(base_data, data))
                ):
                    sub_instances = tuple(
                        sorted(
                            instance_map[int(base_data[i, j, k])]
                            + instance_map[int(data[i, j, k])]
                        )
                    )
                    if sub_instances not in reverse_instance_map:
                        if not group_instances:
                            return None, {}
                        reverse_instance_map[sub_instances] = group_instances.pop()
                        instance_map[reverse_instance_map[sub_instances]] = list(
                            sub_instances
                        )
                    data[i, j, k] = reverse_instance_map[sub_instances]

                base_data = numpy.where(data, data, base_data)

            if group_instances and group_instances[0] <= 255:
                base_img.set_data_dtype(numpy.uint8)
                base_data = numpy.asarray(base_data, dtype=numpy.uint8)
            else:
                base_data = numpy.asarray(base_data, dtype=numpy.uint16)

            if isinstance(base_img, Nifti1Image):
                new_img = Nifti1Image(base_data, base_img.affine, base_img.header)
            else:
                new_img = Nifti2Image(base_data, base_img.affine, base_img.header)

            dirname = os.path.join(config_path(), "temp", str(uuid4()))
            os.makedirs(dirname, exist_ok=True)
            filename = uniquify_path(os.path.join(dirname, "label.nii.gz"))
            nib_save(new_img, filename)

            group_map: Dict[int, List[int]] = {}
            for sub_ids, group_id in reverse_instance_map.items():
                if len(sub_ids) > 1:
                    for sub_id in sub_ids:
                        if sub_id not in group_map:
                            group_map[sub_id] = []
                        group_map[sub_id].append(group_id)

            for delete_file in delete_files:
                os.remove(delete_file)

            return (filename, group_map)

        except Exception as error:  # pylint: disable=broad-except
            log_error(error)
            return None, {}


async def convert_nii_to_rtstruct(
    nifti_files: List[str],
    dicom_series_path: str,
    categories: List[Dict],
    segment_map: TypeSegmentMap,
    semantic_mask: bool,
) -> Tuple[Optional[Any], TypeSegmentMap]:
    """Convert nifti mask to dicom rt-struct."""
    # pylint: disable=too-many-locals, too-many-branches, import-outside-toplevel
    # pylint: disable=too-many-statements, too-many-return-statements
    async with semaphore:
        import numpy  # type: ignore
        from nibabel.loadsave import load as nib_load  # type: ignore
        from nibabel.nifti1 import Nifti1Image  # type: ignore
        from nibabel.nifti2 import Nifti2Image  # type: ignore
        from rt_utils import RTStructBuilder  # type: ignore

        try:
            category_map = {
                category.get("category", ""): (
                    int(category.get("classId", -1)),
                    category.get("color", ""),
                    category.get("parents", []) or [],
                )
                for category in categories
            }
            rtstruct = RTStructBuilder.create_new(dicom_series_path=dicom_series_path)
            rtstruct.ds.InstitutionName = "RedBrick AI"
            rtstruct.ds.Manufacturer = "RedBrick AI"
            rtstruct.ds.ManufacturerModelName = "redbrick-sdk"

            new_segment_map: TypeSegmentMap = {}
            for nifti_file in nifti_files:
                img: Nifti1Image = nib_load(nifti_file)  # type: ignore

                if not isinstance(img, Nifti1Image) and not isinstance(
                    img, Nifti2Image
                ):
                    return None, {}

                # Load NIfTI file
                data = img.get_fdata(caching="unchanged")
                data = numpy.round(data).astype(numpy.uint16)  # type: ignore

                if data.ndim != 3:
                    return None, {}

                data = data.swapaxes(0, 1)
                unique_instances: Set[int] = set(numpy.unique(data))
                if 0 in unique_instances:
                    unique_instances.remove(0)

                cat: Optional[Union[str, Dict]] = None
                category: Optional[str] = None
                if semantic_mask:
                    new_data = numpy.zeros_like(data)
                    segment_remap: Dict[str, Union[str, Dict]] = {}
                    for instance in unique_instances:
                        cat = segment_map.get(str(instance))  # type: ignore
                        category = cat.get("category") if isinstance(cat, dict) else cat
                        if not (
                            category
                            and category_map.get(category)
                            and category_map[category][0] >= 0
                        ):
                            continue

                        new_instance = category_map[category][0] + 1
                        new_data[data == instance] = new_instance
                        if str(new_instance) not in segment_remap:
                            segment_remap[str(new_instance)] = cat  # type: ignore

                    data = new_data
                    segment_map.clear()
                    segment_map.update(segment_remap)  # type: ignore
                    unique_instances.clear()
                    unique_instances.update(
                        {int(new_key) for new_key in segment_map.keys()}
                    )

                for instance in unique_instances:
                    kwargs = {}
                    cat = segment_map.get(str(instance))  # type: ignore
                    category = cat.get("category") if isinstance(cat, dict) else cat
                    roi_name = (
                        category
                        if category and semantic_mask
                        else f"Segment_{instance}"
                    )
                    if cat:
                        new_segment_map[roi_name] = cat  # type: ignore

                    if category and category_map.get(category):
                        selected_cat = category_map[category]
                        kwargs["color"] = selected_cat[1]
                        kwargs["description"] = (
                            f"{selected_cat[0]} - "
                            + "".join((parent + "/") for parent in selected_cat[2])
                            + category
                        )

                    rtstruct.add_roi(mask=data == instance, name=roi_name, **kwargs)

            return rtstruct, new_segment_map
        except Exception as error:  # pylint: disable=broad-except
            log_error(error)
            return None, {}


def merge_rtstructs(rtstruct1: Any, rtstruct2: Any) -> Any:
    """Merge two rtstructs."""
    for roi_contour_seq, struct_set_roi_seq, rt_roi_observation_seq in zip(
        rtstruct1.ds.ROIContourSequence,
        rtstruct1.ds.StructureSetROISequence,
        rtstruct1.ds.RTROIObservationsSequence,
    ):
        roi_name = struct_set_roi_seq.ROIName

        # Check for ROI name duplication in rtstruct2
        duplicate_roi = next(
            (
                roi_seq2
                for roi_seq2 in rtstruct2.ds.StructureSetROISequence
                if roi_seq2.ROIName == roi_name
            ),
            None,
        )

        if duplicate_roi:
            # If ROI name already exists in rtstruct2, append contours to the existing ROI
            duplicate_roi_contour_seq = next(
                (
                    contour_seq
                    for contour_seq in rtstruct2.ds.ROIContourSequence
                    if contour_seq.ReferencedROINumber == duplicate_roi.ROINumber
                ),
                None,
            )
            if duplicate_roi_contour_seq:
                duplicate_roi_contour_seq.ContourSequence.extend(
                    roi_contour_seq.ContourSequence
                )
            else:
                # If no contour sequence found for the duplicate ROI, create a new one
                rtstruct2.ds.ROIContourSequence.append(roi_contour_seq)
                duplicate_roi.ROINumber = len(rtstruct2.ds.StructureSetROISequence)
        else:
            # If ROI name is unique, proceed with existing logic
            roi_number = len(rtstruct2.ds.StructureSetROISequence) + 1
            roi_contour_seq.ReferencedROINumber = roi_number
            struct_set_roi_seq.ROINumber = roi_number
            rt_roi_observation_seq.ReferencedROINumber = roi_number

            rtstruct2.ds.ROIContourSequence.append(roi_contour_seq)
            rtstruct2.ds.StructureSetROISequence.append(struct_set_roi_seq)
            rtstruct2.ds.RTROIObservationsSequence.append(rt_roi_observation_seq)

    return rtstruct2


async def convert_rtstruct_to_nii(
    rt_struct_files: List[str],
    dicom_series_path: str,
    segment_map: TypeSegmentMap,
    label_validate: bool,
    categories: List[Dict],
) -> Tuple[Optional[Any], TypeSegmentMap]:
    """Convert dicom rt-struct to nifti mask."""
    # pylint: disable=too-many-locals, too-many-branches, import-outside-toplevel
    # pylint: disable=too-many-statements, too-many-return-statements
    if not rt_struct_files:
        log_error("No segmentations found")
        return None, {}

    async with semaphore:
        import numpy  # type: ignore
        from nibabel.nifti1 import Nifti1Image  # type: ignore
        from rt_utils import RTStructBuilder  # type: ignore

        rt_struct = RTStructBuilder.create_from(dicom_series_path, rt_struct_files[0])
        for rt_struct_file in rt_struct_files[1:]:
            rt_struct = merge_rtstructs(
                rt_struct,
                RTStructBuilder.create_from(dicom_series_path, rt_struct_file),
            )

        roi_names = set(rt_struct.get_roi_names())

        for name in set(segment_map.keys()):
            if name not in roi_names:
                del segment_map[name]

        misses = roi_names - set(segment_map.keys())  # type: ignore
        if misses:
            logger.warning(
                f"Following ROI names are present in the file, but not in segmentMap: {misses}"
            )

        if label_validate:
            category_names = {cat["category"] for cat in categories}
            for cat in segment_map.values():
                if isinstance(cat, dict):
                    cat = cat.get("category")  # type: ignore
                if cat not in category_names:
                    log_error(
                        f"ROI '{cat}' of labelType SEGMENTATION is not present in the taxonomy"
                    )
                    return None, {}

        new_segment_map: TypeSegmentMap = {}
        dtype = numpy.uint16 if len(roi_names) >= 250 else numpy.uint8

        nii_mask = numpy.zeros(
            (
                rt_struct.series_data[0].Columns,
                rt_struct.series_data[0].Rows,
                len(rt_struct.series_data),
            ),
            dtype,
        )

        for idx, name in enumerate(segment_map.keys()):
            if name not in roi_names:
                continue
            new_segment_map[str(idx + 1)] = segment_map[name]
            roi_mask = rt_struct.get_roi_mask_by_name(name)
            nii_mask[roi_mask] = idx + 1

        return (
            Nifti1Image(nii_mask.swapaxes(0, 1), numpy.diag([1, 2, 3, 1])),
            new_segment_map,
        )



"""Interface for getting basic information about a workspace."""

from typing import Dict
from abc import ABC, abstractmethod


class WorkspaceRepoInterface(ABC):
    """Abstract interface to Workspace APIs."""

    @abstractmethod
    def get_workspace(self, org_id: str, workspace_id: str) -> Dict:
        """
        Get workspace name and status.

        Raise an exception if workspace does not exist.
        """

    @abstractmethod
    def create_workspace(
        self, org_id: str, workspace_name: str, exists_okay: bool = False
    ) -> Dict:
        """
        Create a workspace.

        Raise an exception if workspace already exists.
        """



"""Initialize repo module."""

from .export import ExportRepo
from .labeling import LabelingRepo
from .upload import UploadRepo
from .settings import SettingsRepo
from .project import ProjectRepo
from .workspace import WorkspaceRepo



"""Taxonomy utilities."""

from typing import Dict, List, Optional


def format_taxonomy(taxonomy: Dict) -> Dict:
    """Parse taxonomy object."""
    keys = ["orgId", "name", "createdAt", "archived", "isNew"]
    if taxonomy["isNew"]:
        keys.extend(
            [
                "taxId",
                "studyClassify",
                "seriesClassify",
                "instanceClassify",
                "objectTypes",
            ]
        )
    else:
        keys.extend(
            [
                "version",
                "categories",
                "attributes",
                "taskCategories",
                "taskAttributes",
                "colorMap",
            ]
        )

    return {key: taxonomy[key] for key in keys}


def validate_attribute(attribute: Dict, message: str) -> None:
    """Validate attribute."""
    if attribute.get("name") is None:
        raise ValueError(f"{message} has no `name`")
    if attribute.get("attrType") is None:
        raise ValueError(f"{message} has no `attrType`")
    if attribute.get("attrId") is None:
        raise ValueError(f"{message} has no `attrId`")


def validate_taxonomy(
    study_classify: Optional[List[Dict]],
    series_classify: Optional[List[Dict]],
    instance_classify: Optional[List[Dict]],
    object_types: Optional[List[Dict]],
) -> None:
    """Validate taxonomy."""
    for aidx, attribute in enumerate(study_classify or []):
        validate_attribute(attribute, f"study_classify->{aidx}")

    for aidx, attribute in enumerate(series_classify or []):
        validate_attribute(attribute, f"series_classify->{aidx}")

    for aidx, attribute in enumerate(instance_classify or []):
        validate_attribute(attribute, f"instance_classify->{aidx}")

    for oidx, object_type in enumerate(object_types or []):
        message = f"object_types->{oidx}"
        if object_type.get("category") is None:
            raise ValueError(f"{message} has no `category`")
        if object_type.get("classId") is None:
            raise ValueError(f"{message} has no `classId`")
        if object_type.get("labelType") is None:
            raise ValueError(f"{message} has no `labelType`")

        for aidx, attribute in enumerate(object_type.get("attributes", []) or []):
            validate_attribute(attribute, f"{message}->attributes->{aidx}")



"""CLI init command."""

import os
from argparse import ArgumentParser, Namespace
from typing import List

from rich.console import Console

from redbrick.cli.input import CLIInputNumber, CLIInputSelect, CLIInputText
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIInitInterface
from redbrick.organization import RBOrganization
from redbrick.utils.logging import assert_validation


class CLIInitController(CLIInitInterface):
    """CLI init command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize init sub commands."""
        parser.add_argument("--name", "-n", help="Project name")
        parser.add_argument("--taxonomy", "-t", help="Taxonomy name")
        parser.add_argument(
            "--reviews",
            "-r",
            help="Number of review stages",
        )
        parser.add_argument(
            "path",
            nargs="?",
            default=".",
            help="Local path of the empty project directory",
        )

    def handler(self, args: Namespace) -> None:
        """Handle init command."""
        self.args = args
        project = CLIProject.from_path(path=self.args.path, required=False)
        assert_validation(project is None, "Already inside a project")

        path = os.path.realpath(self.args.path)
        if os.path.exists(path):
            if not os.path.isdir(path):
                raise Exception(f"Not a directory {path}")
            if os.listdir(path):
                raise Exception(f"{path} is not empty")
        else:
            os.makedirs(path)

        self.project = CLIProject(path=self.args.path, required=False)

        self.handle_init()

    def handle_init(self) -> None:
        """Handle empty sub command."""
        assert_validation(self.project.creds.exists, "Credentials missing")

        console = Console()
        with console.status("Fetching organization") as status:
            try:
                org = RBOrganization(self.project.context, self.project.creds.org_id)
            except Exception as error:
                status.stop()
                raise error
        console.print("[bold green]" + str(org))

        with console.status("Fetching taxonomies") as status:
            try:
                taxonomies: List[str] = [
                    taxonomy["name"]  # type: ignore
                    for taxonomy in org.taxonomies(False)
                    if taxonomy.get("isNew") and not taxonomy.get("archived")  # type: ignore
                ]
            except Exception as error:
                status.stop()
                raise error

        name = CLIInputText(
            self.args.name, "Name", os.path.basename(self.project.path)
        ).get()
        taxonomy = CLIInputSelect(self.args.taxonomy, "Taxonomy", taxonomies).get()
        reviews = int(CLIInputNumber(self.args.reviews, "Reviews").get())

        with console.status("Creating project") as status:
            try:
                project = org.create_project(name, taxonomy, reviews)
            except Exception as error:
                status.stop()
                raise error
        console.print("[bold green]" + str(project))

        self.project.initialize_project(org, project)



"""Model stage."""

from dataclasses import dataclass, field
import json
from typing import Any, Dict, List, Optional, TypedDict

from redbrick.common.stage import Stage


class ModelTaxonomyMap(TypedDict):
    """Model taxonomy map."""

    category: str
    classid: int


@dataclass
class ModelStage(Stage):
    """Model Stage.

    Parameters
    --------------
    stage_name: str
        Stage name.

    on_submit: Optional[str] = None
        The next stage for the task when submitted in current stage.
        If None, will go to Output stage.

    config: Config
        Stage config.
    """

    @dataclass
    class Config(Stage.Config):
        """Model Stage Config.

        Parameters
        --------------
        name: str
            Model name.

        url: Optional[str]
            URL for self-hosted model.

        taxonomy_objects: Optional[Dict[str, int]]
            Mapping of model classes to project's taxonomy objects.
        """

        name: str
        url: Optional[str] = None
        taxonomy_objects: Optional[List[ModelTaxonomyMap]] = None

        @classmethod
        def from_entity(cls, entity: Optional[Dict] = None) -> "ModelStage.Config":
            """Get object from entity."""
            if not entity:
                raise ValueError("Model name is required")
            return cls(
                name=entity["name"],
                url=entity.get("url"),
                taxonomy_objects=entity.get("taxonomyObjects"),
            )

        def to_entity(self) -> Dict:
            """Get entity from object."""
            entity: Dict[str, Any] = {"name": self.name}
            if self.url is not None:
                entity["url"] = self.url
            if self.taxonomy_objects is not None:
                entity["taxonomyObjects"] = self.taxonomy_objects
            return entity

    stage_name: str
    on_submit: Optional[str] = None
    config: Config = field(default_factory=Config.from_entity)

    @classmethod
    def from_entity(cls, entity: Dict) -> "ModelStage":
        """Get object from entity"""
        config = entity.get("stageConfig")
        if config and isinstance(config, str):
            config = json.loads(config)
        return cls(
            stage_name=entity["stageName"],
            on_submit=entity["routing"]["nextStageName"],
            config=cls.Config.from_entity(config or {}),
        )

    def to_entity(self) -> Dict:
        """Get entity from object."""
        return {
            "brickName": "model",
            "stageName": self.stage_name,
            "routing": {
                "nextStageName": self.on_submit or "Output",
            },
            "stageConfig": self.config.to_entity(),
        }



"""CLI report command."""

import os
import json
from datetime import datetime
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import cast

from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIReportInterface
from redbrick.utils.logging import assert_validation, logger


class CLIIReportController(CLIReportInterface):
    """CLI report command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize report sub commands."""
        parser.add_argument(
            "type",
            nargs="?",
            default=self.TYPE_ALL,
            help=f"Export type: ({self.TYPE_ALL} [default], {self.TYPE_GROUNDTRUTH})",
        )
        parser.add_argument(
            "--concurrency",
            "-c",
            type=int,
            default=10,
            help="Concurrency value (Default: 10)",
        )

    def handler(self, args: Namespace) -> None:
        """Handle report command."""
        self.args = args
        project = CLIProject.from_path()
        assert_validation(project, "Not a valid project")
        self.project = cast(CLIProject, project)

        self.handle_report()

    def handle_report(self) -> None:
        """Handle empty sub command."""
        if self.args.type not in (self.TYPE_ALL, self.TYPE_GROUNDTRUTH):
            raise ArgumentError(None, f"Invalid report type: {self.args.type}")

        reports = self.project.project.export.get_task_events(
            only_ground_truth=self.args.type == self.TYPE_GROUNDTRUTH,
            concurrency=self.args.concurrency,
        )

        report_file = os.path.abspath(f"report-{int(datetime.now().timestamp())}.json")
        if os.path.isfile(report_file):
            os.remove(report_file)

        for idx, report in enumerate(reports):
            if idx == 0:
                with open(report_file, "wb") as report_file_:
                    report_file_.write(
                        b"[" + json.dumps(report, indent=2).encode("utf-8") + b"]"
                    )
            else:
                with open(report_file, "rb+") as report_file_:
                    report_file_.seek(-1, 2)
                    report_file_.write(
                        b"," + json.dumps(report, indent=2).encode("utf-8") + b"]"
                    )

        if not os.path.isfile(report_file):
            with open(report_file, "w", encoding="utf-8") as report_file_:
                report_file_.write("[]")

        logger.info(f"Exported successfully to: {report_file}")



"""Task types."""

from typing import List, Dict, Literal, Union, TypedDict
from typing_extensions import Required, NotRequired  # type: ignore


class Point2D(TypedDict):
    """Point 2D label."""

    xNorm: float
    yNorm: float


class VoxelPoint(TypedDict):
    """Voxel point coordinates."""

    i: int
    j: int
    k: int


class WorldPoint(TypedDict):
    """World point coordinates."""

    x: float
    y: float
    z: float


class MeasurementStats(TypedDict):
    """Label measurement stats."""

    average: float
    area: NotRequired[float]
    volume: NotRequired[float]
    minimum: float
    maximum: float


class VideoMetaData(TypedDict):
    """Frame and track information for a label."""

    frameIndex: int
    trackId: str
    keyFrame: int
    endTrack: bool


Category = Union[int, str, List[str]]
Attributes = Dict[str, Union[str, bool, List[str]]]


class CommonLabelProps(TypedDict, total=False):
    """Extended version of segmentation mapping."""

    category: Category
    attributes: Attributes
    mask: Union[str, List[str]]


SegmentMap = Dict[Union[str, int], Union[str, int, List[str], CommonLabelProps]]


class InstanceClassification(TypedDict):
    """Instance classification."""

    fileIndex: int
    fileName: NotRequired[str]
    values: Attributes


class Classification(TypedDict):
    """Study or series classification."""

    category: NotRequired[Category]
    attributes: NotRequired[Attributes]
    video: NotRequired[VideoMetaData]


class Polyline(TypedDict):
    """Polyline label."""

    points: List[Point2D]
    category: Category
    attributes: NotRequired[Attributes]
    video: NotRequired[VideoMetaData]


class Polygon(TypedDict):
    """Polygon label."""

    points: List[Point2D]
    category: Category
    attributes: NotRequired[Attributes]
    stats: NotRequired[MeasurementStats]
    video: NotRequired[VideoMetaData]


class Cuboid(TypedDict):
    """Cuboid label."""

    point1: VoxelPoint
    point2: VoxelPoint
    absolutePoint1: WorldPoint
    absolutePoint2: WorldPoint
    category: Category
    attributes: NotRequired[Attributes]
    stats: NotRequired[MeasurementStats]


class BoundingBox(TypedDict):
    """Bounding box label."""

    pointTopLeft: Point2D
    wNorm: float
    hNorm: float
    category: Category
    attributes: NotRequired[Attributes]
    stats: NotRequired[MeasurementStats]
    video: NotRequired[VideoMetaData]


class Ellipse(TypedDict):
    """Ellipse label."""

    pointCenter: Point2D
    xRadiusNorm: float
    yRadiusNorm: float
    rotationRad: float
    category: Category
    attributes: NotRequired[Attributes]
    stats: NotRequired[MeasurementStats]
    video: NotRequired[VideoMetaData]


class MeasureAngle(TypedDict):
    """Angle measurement label."""

    type: Literal["angle"]
    point1: VoxelPoint
    vertex: VoxelPoint
    point2: VoxelPoint
    absolutePoint1: WorldPoint
    absoluteVertex: WorldPoint
    absolutePoint2: WorldPoint
    normal: List[float]
    angle: float
    category: Category
    attributes: NotRequired[Attributes]


class MeasureLength(TypedDict):
    """Length measurement label."""

    type: Literal["length"]
    point1: VoxelPoint
    point2: VoxelPoint
    absolutePoint1: WorldPoint
    absolutePoint2: WorldPoint
    normal: List[float]
    length: float
    category: Category
    attributes: NotRequired[Attributes]


class Landmarks3D(TypedDict):
    """3D point label."""

    point: VoxelPoint
    category: Category
    attributes: NotRequired[Attributes]


class Landmarks(TypedDict):
    """2D point label."""

    point: Point2D
    category: Category
    attributes: NotRequired[Attributes]
    video: NotRequired[VideoMetaData]


class ConsensusScore(TypedDict, total=False):
    """Consensus score (for consensus tasks)."""

    secondaryUser: str
    secondaryUserId: str
    secondaryUserEmail: str
    score: float


class Series(TypedDict, total=False):
    """Task series information."""

    items: Union[str, List[str]]
    name: str
    metaData: Dict[str, str]
    segmentations: Union[str, List[str]]
    segmentMap: SegmentMap
    binaryMask: bool
    semanticMask: bool
    pngMask: bool
    landmarks: List[Landmarks]
    landmarks3d: List[Landmarks3D]
    measurements: List[Union[MeasureLength, MeasureAngle]]
    ellipses: List[Ellipse]
    boundingBoxes: List[BoundingBox]
    cuboids: List[Cuboid]
    polygons: List[Polygon]
    polylines: List[Polyline]
    classifications: List[Classification]
    instanceClassifications: List[InstanceClassification]


class InputTask(TypedDict, total=False):
    """Task object."""

    name: Required[str]
    series: Required[List[Series]]
    classification: Classification

    priority: int
    metaData: Dict[str, str]
    preAssign: Dict[str, Union[str, List[str]]]


class OutputTask(InputTask, total=False):
    """Exported task object."""

    taskId: Required[str]
    currentStageName: str
    status: str

    createdBy: str
    createdAt: str
    updatedBy: str
    updatedByUserId: str
    updatedAt: str

    consensus: bool
    consensusScore: float
    consensusTasks: List["OutputTask"]
    scores: List[ConsensusScore]
    superTruth: "OutputTask"



"""Tests for redbrick.mock_export.public"""
import os
import typing as t
from unittest.mock import patch, Mock, AsyncMock, MagicMock, mock_open

import pytest

import redbrick.export
from tests.fixtures import export as export_fixtures, repo as repo_fixtures


@pytest.mark.unit
def test_get_raw_data_latest(mock_export):
    """Test `redbrick.export.public.Export._get_raw_data_latest`"""
    mock_task_id = "mock_task_id"
    concurrency = 2

    mock_query = Mock(
        return_value=repo_fixtures.get_datapoint_latest_resp(mock_task_id)
    )
    mock_export.context.export.client.execute_query = mock_query
    resp = mock_export._get_raw_data_latest(  # pylint: disable=protected-access
        concurrency, task_id=mock_task_id
    )

    tasks = list(resp)
    assert len(tasks) == 1
    assert tasks[0]["taskId"] == mock_task_id


@pytest.mark.asyncio
@pytest.mark.unit
@pytest.mark.parametrize(
    ("rt_struct", "taxonomy", "check_convert_called"),
    [
        (True, {"taxonomy": []}, False),
        (False, {"isNew": True, "objectTypes": []}, False),
        (True, {"isNew": True, "objectTypes": []}, True),
    ],
)
async def test_download_task_items(
    mock_export, rt_struct, taxonomy, check_convert_called, tmpdir
):
    """Test `redbrick.export.public.Export._download_task_items`"""
    task = export_fixtures.get_tasks_resp[2]
    storage_id = "storage_id"
    parent_dir = str(tmpdir)

    async def mock_download(
        url_path_pairs: t.List[t.Tuple[str, str]], *args, **kwargs
    ):  # pylint: disable=unused-argument
        return [x[1] for x in url_path_pairs]

    mock_convert = AsyncMock(return_value=None)
    mock_export.context.export.presign_items = lambda a, b, items: items
    with patch("redbrick.export.public.download_files", mock_download):
        with patch("redbrick.utils.dicom.convert_nii_to_rtstruct", mock_convert):
            (
                _,
                series_dirs,
            ) = await mock_export._download_task_items(  # pylint: disable=protected-access
                task, storage_id, parent_dir, taxonomy, rt_struct, False
            )
    if check_convert_called:
        mock_convert.assert_called_once()
    assert series_dirs == [
        os.path.join(parent_dir, "BraTS2021_00005", "A"),
        os.path.join(parent_dir, "BraTS2021_00005", "B"),
        os.path.join(parent_dir, "BraTS2021_00005", "C"),
        os.path.join(parent_dir, "BraTS2021_00005", "D"),
    ]
    assert len(series_dirs) == len(task["series"])


@pytest.mark.unit
@pytest.mark.parametrize("get_color_map", [False, True])
def test_preprocess_export(mock_export, get_color_map):
    """Test `redbrick.export.public.Export.preprocess_export`"""
    taxonomy = {
        "isNew": True,
        "objectTypes": [
            {
                "labelType": "SEGMENTATION",
                "classId": "class1",
                "color": "FF0000",
                "category": "Category1",
            }
        ],
    }
    class_map, color_map = mock_export.preprocess_export(
        taxonomy, get_color_map=get_color_map
    )
    if get_color_map:
        assert class_map == {"Category1": [255, 0, 0]}
        assert color_map == {"class1": [255, 0, 0]}
    else:
        assert class_map == color_map == {}


@pytest.mark.unit
@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("task_file", "get_task", "returns_task"),
    [
        (None, False, False),
        (None, True, True),
        ("/tmp/taskfile", True, True),
    ],
)
async def test_export_nifti_label_data(mock_export, task_file, get_task, returns_task):
    """Test `redbrick.export.public.Export.export_nifti_label_data`"""

    # Mock methods
    async def mock_download_task(_task, *args):  # pylint: disable=unused-argument
        return _task

    mock_export._download_task = mock_download_task  # pylint: disable=protected-access
    mock_export.process_labels = AsyncMock(
        return_value=export_fixtures.get_tasks_resp[2]
    )

    datapoint = {
        "labelStorageId": "storage123",
        "labels": ["Label", "Review_1", "END"],
        "items": [],
    }

    taxonomy = {
        "isNew": True,
        "objectTypes": [
            {
                "labelType": "SEGMENTATION",
                "classId": "class1",
                "color": "FF0000",
                "category": "Category1",
            }
        ],
    }
    open_mock = MagicMock(spec=open)
    with patch.object(redbrick.export.public, "open", mock_open(mock=open_mock)):
        task = await mock_export.export_nifti_label_data(
            datapoint,
            taxonomy,
            task_file,
            None,
            None,
            False,
            None,
            False,
            False,
            False,
            False,
            False,
            False,
            get_task,
        )
    mock_export.process_labels.assert_called_once()
    if task_file:
        open_mock.assert_called_once()

    if returns_task:
        assert isinstance(task, dict)
    else:
        assert task is None


@pytest.mark.unit
def test_export_tasks(mock_export, tmpdir):
    """Test `redbrick.export.public.Export.export_tasks`"""
    # Mock the _get_raw_data_latest method
    taxonomy = {
        "isNew": True,
        "objectTypes": [
            {
                "labelType": "SEGMENTATION",
                "classId": "class1",
                "color": "FF0000",
                "category": "Category1",
            }
        ],
    }
    class_map = {"Category1": [255, 0, 0]}
    color_map = {"class1": [255, 0, 0]}
    task_id_to_tasks = {x["taskId"]: x for x in export_fixtures.get_tasks_resp}
    destination_dir = str(tmpdir)

    # patch methods
    mock_export.context.project.get_taxonomy = MagicMock(return_value=taxonomy)
    mock_export.preprocess_export = MagicMock(return_value=(class_map, color_map))
    # patch "self.context.export.get_datapoints_latest" and
    # by extension, "_get_raw_data_latest"
    mock_query = Mock(return_value=repo_fixtures.get_datapoints_latest_resp)
    mock_export.context.export.client.execute_query = mock_query

    async def _mock_nifti(datapoint, *args):  # pylint: disable=unused-argument
        return task_id_to_tasks[datapoint["taskId"]]

    mock_export.export_nifti_label_data = _mock_nifti

    # Mock with_files=True and test
    task_ids = set()
    for task_ in mock_export.export_tasks(destination=destination_dir):
        assert isinstance(task_, dict)
        task_ids.add(task_["taskId"])
    assert task_ids == set(task_id_to_tasks)

    # Mock rt_struct=True and test
    mock_makedirs = MagicMock()
    open_mock = MagicMock(spec=open)
    with patch.object(redbrick.export.public, "open", mock_open(mock=open_mock)):
        with patch.object(redbrick.export.public.os, "makedirs", mock_makedirs):
            task_ = next(
                mock_export.export_tasks(
                    rt_struct=True,
                    with_files=True,
                    without_json=False,
                    without_masks=False,
                    png=True,
                    destination=destination_dir,
                )
            )
    open_mock.assert_called_once()
    assert mock_makedirs.call_count == 2
    assert task_["taskId"] in task_id_to_tasks


@pytest.mark.unit
@pytest.mark.parametrize(
    ("kwargs", "expected_filters", "expected_stage_name"),
    [
        ({"user_id": "mock"}, {"userId"}, "Label"),
        ({"task_id": "mock", "user_id": "mock"}, {"userId", "taskId"}, "Label"),
        ({"user_id": "mock", "search": "ALL"}, {}, None),
        ({"user_id": "mock", "search": "GROUNDTRUTH"}, {}, "END"),
        ({"search": "QUEUED"}, {}, "Label"),
        ({"search": "DRAFT"}, {"status"}, "Label"),
        ({"search": "SKIPPED"}, {"status"}, "Label"),
        ({"search": "COMPLETED"}, {"recentlyCompleted"}, "Label"),
        ({"search": "FAILED", "user_id": "mock"}, {"reviewState"}, "Review_1"),
        ({"search": "ISSUES", "user_id": "mock"}, {"status"}, "Label"),
        ({"search": "BENCHMARK", "user_id": "mock"}, {"benchmark"}, "END"),
        ({"search": "InvalidFilter"}, None, "Label"),
    ],
)
def test_list_tasks(
    mock_export,
    kwargs: t.Dict,
    expected_filters: t.Optional[t.Set],
    expected_stage_name: t.Optional[str],
):
    """
    Test `redbrick.export.public.Export.list_tasks`
    Ensure the right filters gets passed to the Repo method
    """
    # Simulate a call to list_tasks with some parameters
    _tasks = repo_fixtures.task_search_resp("Label")["genericTasks"]["entries"]
    mock_export.context.export.task_search = MagicMock(return_value=(_tasks, None))
    mock_export.context.project.get_members = MagicMock(return_value={})

    search = kwargs.pop("search", "QUEUED")

    if expected_filters is None:
        with pytest.raises(ValueError):
            next(
                mock_export.list_tasks(
                    search=search, limit=10, stage_name="Label", **kwargs
                )
            )
        return

    task = next(
        mock_export.list_tasks(search=search, limit=10, stage_name="Label", **kwargs)
    )

    calls = mock_export.context.export.task_search.mock_calls
    call_args = calls[0].args
    stage_name_ = call_args[2]
    filters = call_args[4]

    assert set(filters) == set(expected_filters)
    assert isinstance(task, dict)
    assert isinstance(task.get("taskId"), str)
    assert stage_name_ == expected_stage_name



"""CLI config command."""

import os
from argparse import ArgumentError, ArgumentParser, Namespace
from typing import List, Optional

from InquirerPy.prompts.confirm import ConfirmPrompt
from rich.console import Console
from rich.table import Table
from rich.box import ROUNDED

from redbrick import _populate_context
from redbrick.cli.input import (
    CLIInputAPIKey,
    CLIInputUUID,
    CLIInputURL,
    CLIInputProfile,
)
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIConfigInterface
from redbrick.common.constants import DEFAULT_URL
from redbrick.common.context import RBContext
from redbrick.organization import RBOrganization
from redbrick.utils.logging import assert_validation


class CLIConfigController(CLIConfigInterface):
    """CLI config command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize config sub commands."""
        sub_command = parser.add_subparsers(dest="sub_command")

        parser.add_argument("--org", "-o", help="Org ID")
        parser.add_argument("--key", "-k", help="API Key")
        parser.add_argument(
            "--url",
            "-u",
            help=f"Endpoint URL, should default to {DEFAULT_URL}.",
        )
        parser.add_argument("--profile", "-p", help="Profile name")
        parser.add_argument(
            "--force",
            "-f",
            action="store_true",
            help="Force create new credentials",
        )

        # list_sub_command
        _ = sub_command.add_parser(
            self.LIST,
            help="List all credential profiles",
            description="List all credential profiles",
        )

        set_sub_command = sub_command.add_parser(
            self.SET,
            help="Set your default credentials profile",
            description="Set your default credentials profile",
        )
        set_sub_command.add_argument("profile", nargs="?", help="Profile name")

        add_sub_command = sub_command.add_parser(
            self.ADD, help="Add a new profile", description="Add a new profile"
        )
        add_sub_command.add_argument("--org", "-o", help="Org ID")
        add_sub_command.add_argument("--key", "-k", help="Add your API Key.")
        add_sub_command.add_argument(
            "--url",
            "-u",
            help=f"Endpoint URL, should default to {DEFAULT_URL}.",
        )
        add_sub_command.add_argument(
            "--profile",
            "-p",
            help="Define a name for your authentication profile.",
        )

        unset_sub_command = sub_command.add_parser(
            self.REMOVE, help="Remove a profile", description="Remove a profile"
        )
        unset_sub_command.add_argument("profile", nargs="?", help="Profile name")

        # clear_sub_command
        _ = sub_command.add_parser(
            self.CLEAR,
            help="Clear all credentials",
            description="Clear all credentials",
        )

        verify_sub_command = sub_command.add_parser(
            self.VERIFY, help="Verify a profile", description="Verify a profile"
        )
        verify_sub_command.add_argument("profile", nargs="?", help="Profile name")

    def handler(self, args: Namespace) -> None:
        """Handle config command."""
        self.args = args
        self.project = CLIProject(required=False)

        if not args.sub_command:
            self.handle_config()
        elif args.sub_command == self.LIST:
            self.handle_list()
        elif args.sub_command == self.SET:
            self.handle_set()
        elif args.sub_command == self.ADD:
            self.handle_add()
        elif args.sub_command == self.REMOVE:
            self.handle_remove()
        elif args.sub_command == self.CLEAR:
            self.handle_clear()
        elif args.sub_command == self.VERIFY:
            self.handle_verify()
        else:
            raise ArgumentError(None, f"Invalid config sub command: {args.sub_command}")

    def handle_config(self) -> None:
        """Handle empty sub command."""
        if self.project.creds.exists:
            if not self.args.force:
                confirmation = ConfirmPrompt(
                    message="Credentials file already exists. Overwrite?", default=False
                ).execute()
                if not confirmation:
                    return
            self.project.creds.remove()

        self.handle_add()

    def handle_list(self) -> None:
        """Handle list sub command."""
        assert_validation(self.project.creds.exists, "Credentials file does not exist")
        default_profile: str = self.project.creds.selected_profile
        profiles: List[str] = self.project.creds.profile_names
        debug_mode = bool(os.environ.get("REDBRICK_SDK_DEBUG"))
        rows: List[List[str]] = []
        table = Table(
            title="[bold green]RedBrick AI Profiles", expand=True, box=ROUNDED
        )
        columns_set = False
        for profile in profiles:
            if not columns_set:
                table.add_column("Name")
            row: List[str] = [profile]
            for key, value in self.project.creds.get_profile(profile).items():
                if not columns_set:
                    table.add_column(
                        key.capitalize(),
                        width=(
                            (43 if debug_mode else 6)
                            if key == "key"
                            else 36 if key == "org" else None
                        ),
                    )
                row.append(
                    ("***" + value[-3:]) if key == "key" and not debug_mode else value
                )
            columns_set = True
            rows.append(row)

        rows.sort(key=lambda profile: [profile[0] == default_profile, profile[0]])

        for name, *row in rows:
            table.add_row(
                ("* " if name == default_profile else "") + name,
                *row,
                style="green" if name == default_profile else None,
            )

        console = Console()
        console.print(table)

    def handle_set(self) -> None:
        """Handle set sub command."""
        assert_validation(self.project.creds.exists, "Credentials file does not exist")
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names
        ).get()
        self.project.creds.set_default(profile)
        self.project.creds.save()

    def handle_add(self) -> None:
        """Handle add sub command."""
        org_id = CLIInputUUID(self.args.org, "Org ID").get()
        api_key = CLIInputAPIKey(self.args.key).get()
        url = CLIInputURL(self.args.url).get()
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names, True
        ).get()

        self.project.creds.add_profile(profile, api_key, org_id, url)
        self.handle_verify(profile)

        self.project.creds.set_default(profile)
        self.project.creds.save()

    def handle_remove(self) -> None:
        """Handle remove sub command."""
        assert_validation(self.project.creds.exists, "Credentials file does not exist")
        profile = CLIInputProfile(
            self.args.profile, self.project.creds.profile_names
        ).get()

        self.project.creds.remove_profile(profile)
        self.project.creds.save()

    def handle_clear(self) -> None:
        """Handle clear sub command."""
        self.project.creds.remove()

    def handle_verify(self, profile: Optional[str] = None) -> None:
        """Handle verify sub command."""
        if profile is None:
            selected_profile = None
            try:
                selected_profile = self.project.creds.selected_profile
            except AssertionError:
                pass
            profile = CLIInputProfile(
                self.args.profile,
                self.project.creds.profile_names,
                default=selected_profile,
            ).get()

        profile_details = self.project.creds.get_profile(profile)
        context = _populate_context(
            RBContext(
                api_key=profile_details["key"].strip(),
                url=profile_details["url"].strip().rstrip("/"),
            )
        )

        console = Console()
        with console.status("Fetching organization") as status:
            try:
                org = RBOrganization(context, profile_details["org"])
            except Exception as error:
                status.stop()
                raise error
        console.print("[bold green]" + str(org))



"""Organization class."""

from datetime import datetime
from functools import partial
from typing import List, Optional, Dict, Union
from tqdm import tqdm  # type: ignore

from redbrick.common.context import RBContext
from redbrick.project import RBProject
from redbrick.workspace import RBWorkspace
from redbrick.stage import Stage, get_project_stages, get_middle_stages

from redbrick.utils.logging import logger
from redbrick.utils.pagination import PaginationIterator
from redbrick.utils.rb_tax_utils import format_taxonomy, validate_taxonomy


class RBOrganization:
    """
    Representation of RedBrick organization.

    The RBOrganization object allows you to programmatically interact with
    your RedBrick organization. This class provides methods for querying your
    organization and doing other high level actions.
    """

    def __init__(self, context: RBContext, org_id: str) -> None:
        """Construct RBOrganization."""
        self.context = context

        self._org_id = org_id
        self._name: str

        self._get_org()

    def _get_org(self) -> None:
        org = self.context.project.get_org(self._org_id)
        self._name = org["name"]

    def taxonomies(self, only_name: bool = True) -> Union[List[str], List[Dict]]:
        """Get a list of taxonomy names/objects in the organization."""
        taxonomies = self.context.project.get_taxonomies(self._org_id)
        if only_name:
            return [tax["name"] for tax in taxonomies]
        return list(map(format_taxonomy, taxonomies))

    def projects_raw(self) -> List[Dict]:
        """Get a list of active projects as raw objects in the organization."""
        projects = self.context.project.get_projects(self._org_id)
        projects = list(filter(lambda x: x["status"] == "CREATION_SUCCESS", projects))

        return projects

    def projects(self) -> List[RBProject]:
        """Get a list of active projects in the organization."""
        projects = self.projects_raw()
        return [
            RBProject(self.context, self._org_id, proj["projectId"])
            for proj in tqdm(projects)
        ]

    @property
    def org_id(self) -> str:
        """Retrieve the unique org_id of this organization."""
        return self._org_id

    @property
    def name(self) -> str:
        """Retrieve unique name of this organization."""
        return self._name

    def __str__(self) -> str:
        """Get string representation of RBOrganization object."""
        return f"RedBrick AI Organization - {self._name} - ( {self._org_id} )"

    def __repr__(self) -> str:
        """Representation of object."""
        return str(self)

    def create_workspace(self, name: str, exists_okay: bool = False) -> RBWorkspace:
        """
        Create a workspace within the organization.

        This method creates an organization in a similar fashion to the
        quickstart on the RedBrick Ai create project page.

        Parameters
        --------------
        name: str
            A unique name for your workspace

        exists_okay: bool = False
            Allow workspaces with the same name to be returned instead of trying to create
            a new workspace. Useful for when running the same script repeatedly when you
            do not want to keep creating new workspaces.

        Returns
        --------------
        redbrick.workspace.RBWorkspace
            A RedBrick Workspace object.
        """
        try:
            workspace_data = self.context.workspace.create_workspace(
                self._org_id, name, exists_okay=exists_okay
            )
        except ValueError as error:
            raise Exception(
                "Project with same name exists, try setting exists_okay=True to"
                + " return this project instead of creating a new one"
            ) from error

        return RBWorkspace(self.context, self._org_id, workspace_data["workspaceId"])

    def create_project_advanced(
        self,
        name: str,
        taxonomy_name: str,
        stages: List[Stage],
        exists_okay: bool = False,
        workspace_id: Optional[str] = None,
    ) -> RBProject:
        """
        Create a project within the organization.

        This method creates an organization in a similar fashion to the
        quickstart on the RedBrick Ai create project page.

        Parameters
        --------------
        name: str
            A unique name for your project

        taxonomy_name: str
            The name of the taxonomy you want to use for this project.
            Taxonomies can be found on the left side bar of the platform.

        stages: List[Stage]
            List of stage configs.

        exists_okay: bool = False
            Allow projects with the same name to be returned instead of trying to create
            a new project. Useful for when running the same script repeatedly when you
            do not want to keep creating new projects.

        workspace_id: Optional[str] = None
            The workspace id that you want to add this project to.

        Returns
        --------------
        redbrick.project.RBProject
            A RedBrick Project object.
        """
        if exists_okay:
            logger.info("exists_okay=True... checking for project with same name")
            all_projects = self.projects_raw()
            same_name = list(filter(lambda x: x["name"] == name, all_projects))
            if same_name:
                temp = RBProject(self.context, self.org_id, same_name[0]["projectId"])
                if temp.td_type != "DICOM_SEGMENTATION":
                    raise ValueError(
                        "Project with matching name exists, but it has a different type"
                    )
                if temp.taxonomy_name != taxonomy_name:
                    raise ValueError(
                        "Project with matching name exists, but it has a different taxonomy"
                    )

                logger.warning(
                    "exists_okay=True... returning project that already existed"
                )
                return temp

        try:
            project_data = self.context.project.create_project(
                self.org_id,
                name,
                get_project_stages(stages),
                "DICOM_SEGMENTATION",
                taxonomy_name,
                workspace_id,
            )
        except ValueError as error:
            raise Exception(
                "Project with same name exists, try setting exists_okay=True to"
                + " return this project instead of creating a new one"
            ) from error

        return RBProject(self.context, self.org_id, project_data["projectId"])

    def create_project(
        self,
        name: str,
        taxonomy_name: str,
        reviews: int = 0,
        exists_okay: bool = False,
        workspace_id: Optional[str] = None,
    ) -> RBProject:
        """
        Create a project within the organization.

        This method creates an organization in a similar fashion to the
        quickstart on the RedBrick Ai create project page.

        Parameters
        --------------
        name: str
            A unique name for your project

        taxonomy_name: str
            The name of the taxonomy you want to use for this project.
            Taxonomies can be found on the left side bar of the platform.

        reviews: int = 0
            The number of review stages that you want to add after the label
            stage.

        exists_okay: bool = False
            Allow projects with the same name to be returned instead of trying to create
            a new project. Useful for when running the same script repeatedly when you
            do not want to keep creating new projects.

        workspace_id: Optional[str] = None
            The workspace id that you want to add this project to.

        Returns
        --------------
        redbrick.project.RBProject
            A RedBrick Project object.
        """
        return self.create_project_advanced(
            name, taxonomy_name, get_middle_stages(reviews), exists_okay, workspace_id
        )

    def get_project(
        self, project_id: Optional[str] = None, name: Optional[str] = None
    ) -> RBProject:
        """Get project by id/name."""
        projects = self.projects_raw()
        if project_id:
            projects = [
                project for project in projects if project["projectId"] == project_id
            ]
        elif name:
            projects = [project for project in projects if project["name"] == name]
        else:
            raise Exception("Either project_id or name is required")

        if not projects:
            raise Exception("No project found")

        return RBProject(self.context, self._org_id, projects[0]["projectId"])

    def labeling_time(
        self, start_date: datetime, end_date: datetime, concurrency: int = 50
    ) -> List[Dict]:
        """Get information of tasks labeled between two dates (both inclusive)."""
        tasks: List[Dict]
        my_iter = PaginationIterator(
            partial(
                self.context.project.get_labeling_information,
                self._org_id,
                start_date,
                end_date,
            ),
            concurrency,
        )
        with tqdm(my_iter, unit=" tasks") as progress:
            tasks = [
                {
                    "orgId": self._org_id,
                    "projectId": task["project"]["projectId"],
                    "taskId": task["taskId"],
                    "completedBy": task["user"]["email"],
                    "timeSpent": task["timeSpent"],
                    "completedAt": task["date"],
                    "cycle": task["cycle"],
                }
                for task in progress
            ]

        return tasks

    def create_taxonomy(
        self,
        name: str,
        study_classify: Optional[List[Dict]] = None,
        series_classify: Optional[List[Dict]] = None,
        instance_classify: Optional[List[Dict]] = None,
        object_types: Optional[List[Dict]] = None,
    ) -> None:
        """
        Create a Taxonomy V2.

        Format reference for categories and attributes objects:
        https://docs.redbrickai.com/python-sdk/format-reference#taxonomy-v2

        Raises
        ----------
        ValueError:
            If there are validation errors.
        """
        validate_taxonomy(
            study_classify, series_classify, instance_classify, object_types
        )
        if self.context.project.create_taxonomy(
            self.org_id,
            name,
            study_classify,
            series_classify,
            instance_classify,
            object_types,
        ):
            logger.info(f"Successfully created taxonomy: {name}")

    def get_taxonomy(
        self, name: Optional[str] = None, tax_id: Optional[str] = None
    ) -> Dict:
        """Get a taxonomy created in your organization based on id or name.

        Format reference for categories and attributes objects:
        https://docs.redbrickai.com/python-sdk/format-reference#taxonomy-v2
        """
        taxonomy = self.context.project.get_taxonomy(self._org_id, tax_id, name)
        return format_taxonomy(taxonomy)

    def update_taxonomy(
        self,
        tax_id: str,
        study_classify: Optional[List[Dict]] = None,
        series_classify: Optional[List[Dict]] = None,
        instance_classify: Optional[List[Dict]] = None,
        object_types: Optional[List[Dict]] = None,
    ) -> None:
        """Update the categories/attributes of Taxonomy (V2) in the organization.

        Format reference for categories and attributes objects:
        https://docs.redbrickai.com/python-sdk/format-reference#taxonomy-v2

        Raises
        ----------
        ValueError:
            If there are validation errors.
        """
        validate_taxonomy(
            study_classify, series_classify, instance_classify, object_types
        )
        if self.context.project.update_taxonomy(
            self._org_id,
            tax_id,
            study_classify,
            series_classify,
            instance_classify,
            object_types,
        ):
            logger.info(f"Successfully updated taxonomy: {tax_id}")



"""Handler for file upload/download."""

import os
import gzip
from typing import Dict, List, Optional, Tuple, Set

import asyncio
import aiohttp
from yarl import URL
from tenacity import Retrying, RetryError
from tenacity.retry import retry_if_not_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random_exponential
from natsort import natsorted, ns

from redbrick.common.constants import MAX_FILE_BATCH_SIZE, MAX_RETRY_ATTEMPTS
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils.logging import log_error, logger


IMAGE_FILE_TYPES = {
    "png": "image/png",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "bmp": "image/bmp",
    "dcm": "application/dicom",
}

VIDEO_FILE_TYPES = {
    "mp4": "video/mp4",
    "avi": "video/x-msvideo",
    "dcm": "application/dicom",
}

DICOM_FILE_TYPES = {
    "": "application/dicom",
    "dcm": "application/dicom",
    "ima": "application/dicom",
    "dicom": "application/dicom",
}

JSON_FILE_TYPES = {"json": "application/json"}

NIFTI_FILE_TYPES = {"nii": "application/octet-stream"}

NRRD_FILE_TYPES = {"nrrd": "application/octet-stream"}

FILE_TYPES = {
    **IMAGE_FILE_TYPES,
    **VIDEO_FILE_TYPES,
    **DICOM_FILE_TYPES,
    **JSON_FILE_TYPES,
    **NIFTI_FILE_TYPES,
    **NRRD_FILE_TYPES,
}

ALL_FILE_TYPES = {"*": "*/*"}


def get_file_type(file_path: str) -> Tuple[str, str]:
    """
    Return file type.

    Return
    ------------
    [file_ext, file_type]
        file_ext: png, jpeg, jpg etc.
        file_type: this is the MIME file type e.g. image/png
    """
    file_path, file_ext = os.path.splitext(file_path.lower())
    if file_ext == ".gz":
        file_path, file_ext = os.path.splitext(file_path)
    file_ext = file_ext.lstrip(".")

    if file_ext not in FILE_TYPES:
        raise ValueError(
            f"Unsupported file type {file_ext}! Only {','.join(FILE_TYPES.keys())} are supported"
        )
    return file_ext, FILE_TYPES[file_ext]


def find_files_recursive(
    root: str, file_types: Set[str], multiple: bool = False
) -> List[List[str]]:
    """Find files recursively in a directory, that belong to a list of allowed file types."""
    if not os.path.isdir(root):
        return []

    items = []
    list_items = []
    discard_list_items = False

    for item in os.listdir(root):
        if item.startswith("."):
            continue
        path = os.path.join(root, item)
        if os.path.isdir(path):
            items.extend(find_files_recursive(path, file_types, multiple))
            discard_list_items = True
        elif os.path.isfile(path) and (  # pylint: disable=too-many-boolean-expressions
            "*" in file_types
            or (
                item.rsplit(".", 1)[-1].lower() in file_types
                or (
                    "." in item
                    and item.rsplit(".", 1)[-1].lower() == "gz"
                    and item.rsplit(".", 2)[-2].lower() in file_types
                )
            )
        ):
            if multiple:
                if not discard_list_items:
                    list_items.append(path)
            else:
                items.append([path])
        else:
            discard_list_items = True

    if (
        not discard_list_items
        and len({item.rsplit(".", 1)[-1].lower() for item in list_items}) == 1
    ):
        items.append(list(natsorted(list_items, alg=ns.IGNORECASE)))  # type: ignore

    return items


def uniquify_path(path: str) -> str:
    """Provide unique path with number index."""
    filename, extension = os.path.splitext(path)
    if extension == ".gz":
        filename, extension = os.path.splitext(filename)
        extension += ".gz"

    counter = 1

    while os.path.exists(path):
        path = filename + " (" + str(counter) + ")" + extension
        counter += 1

    return path


def is_gzipped_data(data: bytes) -> bool:
    """Check if data is gzipped."""
    return data[:2] == b"\x1f\x8b"


def is_dicom_file(file_name: str) -> bool:
    """Check if data is dicom."""
    with open(file_name, "rb") as fp_:
        data = fp_.read()

    if is_gzipped_data(data):
        data = gzip.decompress(data)

    return data[128:132] == b"\x44\x49\x43\x4d"


async def upload_files(
    files: List[Tuple[str, str, str]],
    progress_bar_name: Optional[str] = "Uploading files",
    keep_progress_bar: bool = True,
    verify_ssl: bool = True,
) -> List[bool]:
    """Upload files from local path to url (file path, presigned url, file type)."""

    async def _upload_file(
        session: aiohttp.ClientSession, path: str, url: str, file_type: str
    ) -> bool:
        if not path or not url or not file_type:
            return False

        with open(path, mode="rb") as file_:
            data = file_.read()

        status: int = 0

        headers = {"Content-Type": file_type}
        if not is_gzipped_data(data) and file_type != DICOM_FILE_TYPES[""]:
            headers["Content-Encoding"] = "gzip"
            data = gzip.compress(data)

        try:
            for attempt in Retrying(
                reraise=True,
                stop=stop_after_attempt(MAX_RETRY_ATTEMPTS),
                wait=wait_random_exponential(min=5, max=30),
                retry=retry_if_not_exception_type(KeyboardInterrupt),
            ):
                with attempt:
                    async with session.put(
                        url,
                        headers=headers,
                        data=data,
                        ssl=None if verify_ssl else False,
                    ) as response:
                        status = response.status
        except RetryError as error:
            raise Exception("Unknown problem occurred") from error

        if status == 200:
            return True
        raise ConnectionError(f"Error in uploading {path} to RedBrick")

    conn = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [
            _upload_file(session, path, url, file_type)
            for path, url, file_type in files
        ]
        uploaded = await gather_with_concurrency(
            MAX_FILE_BATCH_SIZE, coros, progress_bar_name, keep_progress_bar
        )

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return uploaded


async def download_files(
    files: List[Tuple[Optional[str], Optional[str]]],
    progress_bar_name: Optional[str] = "Downloading files",
    keep_progress_bar: bool = True,
    overwrite: bool = False,
    zipped: bool = False,
    verify_ssl: bool = True,
) -> List[Optional[str]]:
    """Download files from url to local path (presigned url, file path)."""

    async def _download_file(
        session: aiohttp.ClientSession, url: Optional[str], path: Optional[str]
    ) -> Optional[str]:
        # pylint: disable=no-member
        if not url or not path:
            logger.debug(f"Downloading empty '{url}' to '{path}'")
            return None

        if not overwrite and os.path.isfile(path):
            return path

        headers: Dict = {}
        data: Optional[bytes] = None

        try:
            for attempt in Retrying(
                reraise=True,
                stop=stop_after_attempt(MAX_RETRY_ATTEMPTS),
                wait=wait_random_exponential(min=5, max=30),
                retry=retry_if_not_exception_type(KeyboardInterrupt),
            ):
                with attempt:
                    async with session.get(
                        URL(url, encoded=True),
                        ssl=None if verify_ssl else False,
                    ) as response:
                        if response.status == 200:
                            headers = dict(response.headers)
                            data = await response.read()
        except RetryError as error:
            log_error(error)
            raise Exception("Unknown problem occurred") from error

        if not data:
            logger.debug(f"Received empty data from '{url}'")
            return None

        if not zipped and headers.get("Content-Encoding") == "gzip":
            try:
                data = gzip.decompress(data)
            except Exception:  # pylint: disable=broad-except
                pass
        if zipped and not is_gzipped_data(data):
            data = gzip.compress(data)
        if zipped and not path.endswith(".gz"):
            path += ".gz"
        if not overwrite:
            path = uniquify_path(path)
        with open(path, "wb") as file_:
            file_.write(data)
        return path

    dirs: Set[str] = set()
    for _, path in files:
        if not path:
            continue
        parent = os.path.dirname(path)
        if parent in dirs:
            continue
        if not os.path.isdir(parent):
            logger.debug(f"Creating parent dir for {path}")
            os.makedirs(parent, exist_ok=True)
        dirs.add(parent)

    conn = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=conn) as session:
        coros = [_download_file(session, url, path) for url, path in files]
        paths = await gather_with_concurrency(
            MAX_FILE_BATCH_SIZE,
            coros,
            progress_bar_name,
            keep_progress_bar,
            True,
        )
        await session.close()

    await asyncio.sleep(0.250)  # give time to close ssl connections
    return [(path if isinstance(path, str) else None) for path in paths]
